# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains the available built-in discrete-variable
quantum operations supported by PennyLane, as well as their conventions.
"""
import numpy as np

from pennylane.operation import All, Any, Observable, Operation


class Hadamard(Observable, Operation):
    r"""Hadamard(wires)
    The Hadamard operator

    .. math:: H = \frac{1}{\sqrt{2}}\begin{bmatrix} 1 & 1\\ 1 & -1\end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 0
    num_wires = 1
    par_domain = None


class PauliX(Observable, Operation):
    r"""PauliX(wires)
    The Pauli X operator

    .. math:: \sigma_x = \begin{bmatrix} 0 & 1 \\ 1 & 0\end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 0
    num_wires = 1
    par_domain = None


class PauliY(Observable, Operation):
    r"""PauliY(wires)
    The Pauli Y operator

    .. math:: \sigma_y = \begin{bmatrix} 0 & -i \\ i & 0\end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 0
    num_wires = 1
    par_domain = None


class PauliZ(Observable, Operation):
    r"""PauliZ(wires)
    The Pauli Z operator

    .. math:: \sigma_z = \begin{bmatrix} 1 & 0 \\ 0 & -1\end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 0
    num_wires = 1
    par_domain = None


class S(Operation):
    r"""S(wires)
    The single-qubit phase gate

    .. math:: S = \begin{bmatrix}
                1 & 0 \\
                0 & i
            \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 0
    num_wires = 1
    par_domain = None


class T(Operation):
    r"""T(wires)
    The single-qubit T gate

    .. math:: T = \begin{bmatrix}
                1 & 0 \\
                0 & e^{\frac{i\pi}{4}}
            \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 0
    num_wires = 1
    par_domain = None


class CNOT(Operation):
    r"""CNOT(wires)
    The controlled-NOT operator

    .. math:: CNOT = \begin{bmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0\\
            0 & 0 & 0 & 1\\
            0 & 0 & 1 & 0
        \end{bmatrix}.

    .. note:: The first wire provided corresponds to the **control qubit**.

    **Details:**

    * Number of wires: 2
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wires the operation acts on
    """
    num_params = 0
    num_wires = 2
    par_domain = None


class CZ(Operation):
    r"""CZ(wires)
    The controlled-Z operator

    .. math:: CZ = \begin{bmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0\\
            0 & 0 & 1 & 0\\
            0 & 0 & 0 & -1
        \end{bmatrix}.

    .. note:: The first wire provided corresponds to the **control qubit**.

    **Details:**

    * Number of wires: 2
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wires the operation acts on
    """
    num_params = 0
    num_wires = 2
    par_domain = None


class SWAP(Operation):
    r"""SWAP(wires)
    The swap operator

    .. math:: SWAP = \begin{bmatrix}
            1 & 0 & 0 & 0 \\
            0 & 0 & 1 & 0\\
            0 & 1 & 0 & 0\\
            0 & 0 & 0 & 1
        \end{bmatrix}.

    **Details:**

    * Number of wires: 2
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wires the operation acts on
    """
    num_params = 0
    num_wires = 2
    par_domain = None

class CSWAP(Operation):
    r"""CSWAP(wires)
    The controlled-swap operator

    .. math:: CSWAP = \begin{bmatrix}
            1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 \\
            0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 1
        \end{bmatrix}.

    .. note:: The first wire provided corresponds to the **control qubit**.

    **Details:**

    * Number of wires: 3
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wires the operation acts on
    """
    num_params = 0
    num_wires = 3
    par_domain = None


class RX(Operation):
    r"""RX(phi, wires)
    The single qubit X rotation

    .. math:: R_x(\phi) = e^{-i\phi\sigma_x/2} = \begin{bmatrix}
                \cos(\phi/2) & -i\sin(\phi/2) \\
                -i\sin(\phi/2) & \cos(\phi/2)
            \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 1
    * Gradient recipe: :math:`\frac{d}{d\phi}f(R_x(\phi)) = \frac{1}{2}\left[f(R_x(\phi+\pi/2)) - f(R_x(\phi-\pi/2))\right]`
      where :math:`f` is an expectation value depending on :math:`R_x(\phi)`.

    Args:
        phi (float): rotation angle :math:`\phi`
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 1
    num_wires = 1
    par_domain = "R"
    grad_method = "A"
    generator = [PauliX, -1/2]


class RY(Operation):
    r"""RY(phi, wires)
    The single qubit Y rotation

    .. math:: R_y(\phi) = e^{-i\phi\sigma_y/2} = \begin{bmatrix}
                \cos(\phi/2) & -\sin(\phi/2) \\
                \sin(\phi/2) & \cos(\phi/2)
            \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 1
    * Gradient recipe: :math:`\frac{d}{d\phi}f(R_y(\phi)) = \frac{1}{2}\left[f(R_y(\phi+\pi/2)) - f(R_y(\phi-\pi/2))\right]`
      where :math:`f` is an expectation value depending on :math:`R_y(\phi)`.

    Args:
        phi (float): rotation angle :math:`\phi`
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 1
    num_wires = 1
    par_domain = "R"
    grad_method = "A"
    generator = [PauliY, -1/2]


class RZ(Operation):
    r"""RZ(phi, wires)
    The single qubit Z rotation

    .. math:: R_z(\phi) = e^{-i\phi\sigma_z/2} = \begin{bmatrix}
                e^{-i\phi/2} & 0 \\
                0 & e^{i\phi/2}
            \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 1
    * Gradient recipe: :math:`\frac{d}{d\phi}f(R_z(\phi)) = \frac{1}{2}\left[f(R_z(\phi+\pi/2)) - f(R_z(\phi-\pi/2))\right]`
      where :math:`f` is an expectation value depending on :math:`R_z(\phi)`.

    Args:
        phi (float): rotation angle :math:`\phi`
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 1
    num_wires = 1
    par_domain = "R"
    grad_method = "A"
    generator = [PauliZ, -1/2]


class PhaseShift(Operation):
    r"""PhaseShift(phi, wires)
    Arbitrary single qubit local phase shift

    .. math:: R_\phi(\phi) = e^{i\phi/2}R_z(\phi) = \begin{bmatrix}
                1 & 0 \\
                0 & e^{i\phi}
            \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 1
    * Gradient recipe: :math:`\frac{d}{d\phi}f(R_\phi(\phi)) = \frac{1}{2}\left[f(R_\phi(\phi+\pi/2)) - f(R_\phi(\phi-\pi/2))\right]`
      where :math:`f` is an expectation value depending on :math:`R_{\phi}(\phi)`.

    Args:
        phi (float): rotation angle :math:`\phi`
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 1
    num_wires = 1
    par_domain = "R"
    grad_method = "A"
    generator = [np.array([[0, 0], [0, 1]]), 1]


class Rot(Operation):
    r"""Rot(phi, theta, omega, wires)
    Arbitrary single qubit rotation

    .. math:: R(\phi,\theta,\omega) = RZ(\omega)RY(\theta)RZ(\phi)= \begin{bmatrix}
                e^{-i(\phi+\omega)/2}\cos(\theta/2) & -e^{i(\phi-\omega)/2}\sin(\theta/2) \\
                e^{-i(\phi-\omega)/2}\sin(\theta/2) & e^{i(\phi+\omega)/2}\cos(\theta/2)
            \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 3
    * Gradient recipe: :math:`\frac{d}{d\phi}f(R(\phi, \theta, \omega)) = \frac{1}{2}\left[f(R(\phi+\pi/2, \theta, \omega)) - f(R(\phi-\pi/2, \theta, \omega))\right]`
      where :math:`f` is an expectation value depending on :math:`R(\phi, \theta, \omega)`.
      This gradient recipe applies for each angle argument :math:`\{\phi, \theta, \omega\}`.

    Args:
        phi (float): rotation angle :math:`\phi`
        theta (float): rotation angle :math:`\theta`
        omega (float): rotation angle :math:`\omega`
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 3
    num_wires = 1
    par_domain = "R"
    grad_method = "A"


class CRX(Operation):
    r"""CRX(phi, wires)
    The controlled-RX operator

    .. math:: CR_x(\phi) = \begin{bmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0\\
            0 & 0 & \cos(\phi/2) & -i\sin(\phi/2)\\
            0 & 0 & -i\sin(\phi/2) & \cos(\phi/2)
        \end{bmatrix}.

    .. note:: The first wire provided corresponds to the **control qubit**.

    **Details:**

    * Number of wires: 2
    * Number of parameters: 1
    * Gradient recipe: :math:`\frac{d}{d\phi}f(CR_x(\phi)) = \frac{1}{2}\left[f(CR_x(\phi+\pi/2)) - f(CR_x(\phi-\pi/2))\right]`
      where :math:`f` is an expectation value depending on :math:`CR_x(\phi)`.

    Args:
        phi (float): rotation angle :math:`\phi`
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 1
    num_wires = 2
    par_domain = "R"
    grad_method = "A"
    generator = [np.array([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]]), -1/2]


class CRY(Operation):
    r"""CRY(phi, wires)
    The controlled-RY operator

    .. math:: CR_y(\phi) = \begin{bmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0\\
            0 & 0 & \cos(\phi/2) & -\sin(\phi/2)\\
            0 & 0 & \sin(\phi/2) & \cos(\phi/2)
        \end{bmatrix}.

    .. note:: The first wire provided corresponds to the **control qubit**.

    **Details:**

    * Number of wires: 2
    * Number of parameters: 1
    * Gradient recipe: :math:`\frac{d}{d\phi}f(CR_y(\phi)) = \frac{1}{2}\left[f(CR_y(\phi+\pi/2)) - f(CR_y(\phi-\pi/2))\right]`
      where :math:`f` is an expectation value depending on :math:`CR_y(\phi)`.

    Args:
        phi (float): rotation angle :math:`\phi`
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 1
    num_wires = 2
    par_domain = "R"
    grad_method = "A"
    generator = [np.array([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, -1j], [0, 0, 1j, 0]]), -1/2]


class CRZ(Operation):
    r"""CRZ(phi, wires)
    The controlled-RZ operator

    .. math:: CR_z(\phi) = \begin{bmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0\\
            0 & 0 & e^{-i\phi/2} & 0\\
            0 & 0 & 0 & e^{i\phi/2}
        \end{bmatrix}.

    .. note:: The first wire provided corresponds to the **control qubit**.

    **Details:**

    * Number of wires: 2
    * Number of parameters: 1
    * Gradient recipe: :math:`\frac{d}{d\phi}f(CR_z(\phi)) = \frac{1}{2}\left[f(CR_z(\phi+\pi/2)) - f(CR_z(\phi-\pi/2))\right]`
      where :math:`f` is an expectation value depending on :math:`CR_z(\phi)`.

    Args:
        phi (float): rotation angle :math:`\phi`
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 1
    num_wires = 2
    par_domain = "R"
    grad_method = "A"
    generator = [np.array([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]]), -1/2]


class CRot(Operation):
    r"""CRot(phi, theta, omega, wires)
    The controlled-Rot operator

    .. math:: CR(\phi, \theta, \omega) = \begin{bmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0\\
            0 & 0 & e^{-i(\phi+\omega)/2}\cos(\theta/2) & -e^{i(\phi-\omega)/2}\sin(\theta/2)\\
            0 & 0 & e^{-i(\phi-\omega)/2}\sin(\theta/2) & e^{i(\phi+\omega)/2}\cos(\theta/2)
        \end{bmatrix}.

    .. note:: The first wire provided corresponds to the **control qubit**.

    **Details:**

    * Number of wires: 2
    * Number of parameters: 3
    * Gradient recipe: :math:`\frac{d}{d\phi}f(CR(\phi, \theta, \omega)) = \frac{1}{2}\left[f(CR(\phi+\pi/2, \theta, \omega)) - f(CR(\phi-\pi/2, \theta, \omega))\right]`
      where :math:`f` is an expectation value depending on :math:`CR(\phi, \theta, \omega)`.
      This gradient recipe applies for each angle argument :math:`\{\phi, \theta, \omega\}`.

    Args:
        phi (float): rotation angle :math:`\phi`
        theta (float): rotation angle :math:`\theta`
        omega (float): rotation angle :math:`\omega`
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 3
    num_wires = 2
    par_domain = "R"
    grad_method = "A"


# =============================================================================
# Arbitrary operations
# =============================================================================


class QubitUnitary(Operation):
    r"""QubitUnitary(U, wires)
    Apply an arbitrary fixed unitary matrix.

    **Details:**

    * Number of wires: The operation can act on any number of wires.
    * Number of parameters: 1
    * Gradient recipe: None

    Args:
        U (array[complex]): square unitary matrix
        wires (Sequence[int] or int): the wire(s) the operation acts on
    """
    num_params = 1
    num_wires = Any
    par_domain = "A"
    grad_method = None


# =============================================================================
# State preparation
# =============================================================================


class BasisState(Operation):
    r"""BasisState(n, wires)
    Prepares a single computational basis state.

    **Details:**

    * Number of wires: All (applied to the entire system)
    * Number of parameters: 1
    * Gradient recipe: None (integer parameters not supported)

    Args:
        n (array): prepares the basis state :math:`\ket{n}`, where ``n`` is an
            array of integers from the set :math:`\{0, 1\}`, i.e.,
            if ``n = np.array([0, 1, 0])``, prepares the state :math:`|010\rangle`.
        wires (Sequence[int] or int): the wire(s) the operation acts on
    """
    num_params = 1
    num_wires = Any
    par_domain = "A"
    grad_method = None


class QubitStateVector(Operation):
    r"""QubitStateVector(state, wires)
    Prepare subsystems using the given ket vector in the computational basis.

    **Details:**

    * Number of wires: All (applied to the entire system)
    * Number of parameters: 1
    * Gradient recipe: None

    Args:
        state (array[complex]): a state vector of size 2**len(wires)
        wires (Sequence[int] or int): the wire(s) the operation acts on
    """
    num_params = 1
    num_wires = All
    par_domain = "A"
    grad_method = None


# =============================================================================
# Observables
# =============================================================================


class Hermitian(Observable):
    r"""Hermitian(A, wires)
    An arbitrary Hermitian observable.

    For a Hermitian matrix :math:`A`, the expectation command returns the value

    .. math::
        \braket{A} = \braketT{\psi}{\cdots \otimes I\otimes A\otimes I\cdots}{\psi}

    where :math:`A` acts on the requested wires.

    If acting on :math:`N` wires, then the matrix :math:`A` must be of size
    :math:`2^N\times 2^N`.

    **Details:**

    * Number of wires: Any
    * Number of parameters: 1
    * Gradient recipe: None

    Args:
        A (array): square hermitian matrix
        wires (Sequence[int] or int): the wire(s) the operation acts on
    """
    num_wires = Any
    num_params = 1
    par_domain = "A"
    grad_method = "F"


ops = {
    "Hadamard",
    "PauliX",
    "PauliY",
    "PauliZ",
    "S",
    "T",
    "CNOT",
    "CZ",
    "SWAP",
    "CSWAP",
    "RX",
    "RY",
    "RZ",
    "PhaseShift",
    "Rot",
    "CRX",
    "CRY",
    "CRZ",
    "CRot",
    "BasisState",
    "QubitStateVector",
    "QubitUnitary",
}


obs = {"Hadamard", "PauliX", "PauliY", "PauliZ", "Hermitian"}


__all__ = list(ops | obs)



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane` configuration classe :class:`Configuration`.
"""
import pytest
import os
import logging as log

import toml

import pennylane as qml
from pennylane import Configuration

log.getLogger('defaults')

config_path = 'default_config.toml'

@pytest.fixture(scope="function")
def default_config():
    return Configuration(name=config_path)

@pytest.fixture(scope="session")
def default_config_toml():
    return toml.load(config_path)

class TestConfigurationFileInteraction:
    """Test the interaction with the configuration file."""

    def test_loading_current_directory(self, monkeypatch, default_config_toml):
        """Test that the default configuration file can be loaded
        from the current directory."""

        monkeypatch.chdir(".")
        monkeypatch.setenv("PENNYLANE_CONF", "")
        config = Configuration(name=config_path)

        assert config.path == os.path.join(os.curdir, config_path)
        assert config._config == default_config_toml

    def test_loading_environment_variable(self, monkeypatch, default_config_toml):
        """Test that the default configuration file can be loaded
        from an environment variable."""

        os.curdir = "None"
        monkeypatch.setenv("PENNYLANE_CONF", os.getcwd())

        config = Configuration(name=config_path)

        assert config._config == default_config_toml
        assert config._env_config_dir == os.environ["PENNYLANE_CONF"]
        assert config.path == os.path.join(os.environ["PENNYLANE_CONF"], config_path)

    def test_loading_absolute_path(self, monkeypatch, default_config_toml):
        """Test that the default configuration file can be loaded
        from an absolute path."""

        os.curdir = "None"
        monkeypatch.setenv("PENNYLANE_CONF", "")

        config = Configuration(name=os.path.join(os.getcwd(), config_path))

        assert config._config == default_config_toml
        assert config.path == os.path.join(os.getcwd(), config_path)

    def test_not_found_warning(self, caplog):
        """Test that a warning is raised if no configuration file found."""

        caplog.clear()
        caplog.set_level(log.INFO)

        Configuration("noconfig")

        assert len(caplog.records) == 1
        assert caplog.records[0].message == "No PennyLane configuration file found."

    def test_save(self, tmp_path):
        """Test saving a configuration file."""
        config = Configuration(name=config_path)

        # make a change
        config['strawberryfields.global']['shots'] = 10

        # Need to convert to string for Python 3.5 compatibility
        temp_config_path = str(tmp_path / 'test_config.toml')
        config.save(temp_config_path)

        result = toml.load(temp_config_path)
        config._config == result

class TestProperties:
    """Test that the configuration class works as expected"""

    def test_get_item(self, default_config):
        """Test getting items."""
        # get existing options
        assert default_config['main.shots'] == 1000
        assert default_config['main']['shots'] == 1000
        assert default_config['strawberryfields.global.hbar'] == 1
        assert default_config['strawberryfields.global']['hbar'] == 1

        # get nested dictionaries
        assert default_config['strawberryfields.fock'] == {'cutoff_dim': 10}

        # get key that doesn't exist
        assert default_config['projectq.ibm.idonotexist'] == {}

    def test_set_item(self, default_config):
        """Test setting items."""

        # set existing options
        default_config['main.shots'] = 10
        assert default_config['main.shots'] == 10
        assert default_config['main']['shots'] == 10

        default_config['strawberryfields.global']['hbar'] = 5
        assert default_config['strawberryfields.global.hbar'] == 5
        assert default_config['strawberryfields.global']['hbar'] == 5

        # set new options
        default_config['projectq.ibm']['device'] = 'ibmqx4'
        assert default_config['projectq.ibm.device'] == 'ibmqx4'

        # set nested dictionaries
        default_config['strawberryfields.tf'] = {'batched': True, 'cutoff_dim': 6}
        assert default_config['strawberryfields.tf'] == {'batched': True, 'cutoff_dim': 6}

        # set nested keys that don't exist dictionaries
        default_config['strawberryfields.another.hello.world'] = 5
        assert default_config['strawberryfields.another'] == {'hello': {'world': 5}}

    def test_bool(self, default_config):
        """Test boolean value of the Configuration object."""

        # test false if no config is loaded
        config = Configuration('noconfig')

        assert not config
        assert default_config

class TestPennyLaneInit:
    """Tests to ensure that the code in PennyLane/__init__.py
    correctly knows how to load and use configuration data"""

    def test_device_load(self, default_config):
        """Test loading a device with a configuration."""
        dev = qml.device('default.gaussian', wires=2, config=default_config)

        assert dev.hbar == 2
        assert dev.shots == 1000



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This is the top level module from which all basic functions and classes of
PennyLane can be directly imported.
"""
import os
from pkg_resources import iter_entry_points

from autograd import numpy
from autograd import grad as _grad
from autograd import jacobian as _jacobian

from semantic_version import Version, Spec

import pennylane.operation

import pennylane.init
import pennylane.templates.layers
import pennylane.templates.embeddings
from pennylane.about import about


from .circuit_graph import CircuitGraph
from .configuration import Configuration
from ._device import Device, DeviceError
from .measure import expval, var, sample
from .ops import *
from .optimize import *
from .qnode import QNode, QuantumFunctionError
from ._version import __version__

# NOTE: this has to be imported last,
# otherwise it will clash with the .qnode import.
from .decorator import qnode


# overwrite module docstrings
numpy.__doc__ = "NumPy with automatic differentiation support, provided by Autograd."


# Look for an existing configuration file
default_config = Configuration("config.toml")


# get list of installed plugin devices
plugin_devices = {entry.name: entry for entry in iter_entry_points("pennylane.plugins")}


def device(name, *args, **kwargs):
    r"""device(name, wires=1, *args, **kwargs)
    Load a plugin :class:`~.Device` and return the instance.

    This function is used to load a particular quantum device,
    which can then be used to construct QNodes.

    PennyLane comes with support for the following two devices:

    * :mod:`'default.qubit' <pennylane.plugins.default_qubit>`: a simple pure
      state simulator of qubit-based quantum circuit architectures.

    * :mod:`'default.gaussian' <pennylane.plugins.default_gaussian>`: a simple simulator
      of Gaussian states and operations on continuous-variable circuit architectures.

    In addition, additional devices are supported through plugins — see
    the  `available plugins <https://pennylane.ai/plugins.html>`_ for more
    details.

    All devices must be loaded by specifying their **short-name** as listed above,
    followed by the number of *wires* (subsystems) you wish to initialize.

    Some devices may accept additional arguments. For instance,
    ``default.gaussian`` accepts the keyword argument ``hbar``, to set
    the convention used in the commutation relation :math:`[\x,\p]=i\hbar`
    (by default set to 2).

    Please refer to the documentation for the individual devices to see any
    additional arguments that might be required or supported.

    Args:
        name (str): the name of the device to load
        wires (int): the number of wires (subsystems) to initialise
            the device with

    Keyword Args:
        config (pennylane.Configuration): a PennyLane configuration object
            that contains global and/or device specific configurations.
    """
    if name in plugin_devices:
        options = {}

        # load global configuration settings if available
        config = kwargs.get("config", default_config)

        if config:
            # combine configuration options with keyword arguments.
            # Keyword arguments take preference, followed by device options,
            # followed by plugin options, followed by global options.
            options.update(config["main"])
            options.update(config[name.split(".")[0] + ".global"])
            options.update(config[name])

        kwargs.pop("config", None)
        options.update(kwargs)

        # loads the plugin device class
        plugin_device_class = plugin_devices[name].load()

        if Version(version()) not in Spec(plugin_device_class.pennylane_requires):
            raise DeviceError(
                "The {} plugin requires PennyLane versions {}, however PennyLane "
                "version {} is installed.".format(
                    name, plugin_device_class.pennylane_requires, __version__
                )
            )

        # load plugin device
        return plugin_device_class(*args, **options)

    raise DeviceError(
        "Device does not exist. Make sure the required plugin is installed."
    )


def grad(func, argnum):
    """Returns the gradient as a callable function of (functions of) QNodes.

    This is a wrapper around the :mod:`autograd.grad` functions.

    Args:
        func (function): a Python function or QNode that contains
            a combination of quantum and classical nodes
        argnum (int or list(int)): which argument(s) to take the gradient
            with respect to

    Returns:
        function: the function that returns the gradient of the input
        function with respect to the arguments in argnum
    """
    # pylint: disable=no-value-for-parameter
    return _grad(func, argnum)


def jacobian(func, argnum):
    """Returns the Jacobian as a callable function of vector-valued
    (functions of) QNodes.

    This is a wrapper around the :mod:`autograd.jacobian` function.

    Args:
        func (function): a vector-valued Python function or QNode that contains
            a combination of quantum and classical nodes. The output of the computation
            must consist of a single NumPy array (if classical) or a tuple of
            expectation values (if a quantum node)
        argnum (int or Sequence[int]): which argument to take the gradient
            with respect to. If a sequence is given, the Jacobian matrix
            corresponding to all input elements and all output elements is returned.

    Returns:
        function: the function that returns the Jacobian of the input
        function with respect to the arguments in argnum
    """
    # pylint: disable=no-value-for-parameter
    if isinstance(argnum, int):
        return _jacobian(func, argnum)
    return lambda *args, **kwargs: numpy.stack([_jacobian(func, arg)(*args, **kwargs) for arg in argnum]).T


def version():
    """Returns the PennyLane version number."""
    return __version__



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains the :func:`TFQNode` function to convert Numpy-interfacing quantum nodes to TensorFlow
compatible quantum nodes.
"""
# pylint: disable=redefined-outer-name
from functools import partial

import numpy as np
import tensorflow as tf

from pennylane.utils import unflatten


if tf.__version__[0] == "1":
    import tensorflow.contrib.eager as tfe # pylint: disable=unused-import,ungrouped-imports
    Variable = tfe.Variable
else:
    from tensorflow import Variable # pylint: disable=unused-import,ungrouped-imports


def TFQNode(qnode):
    """Function that accepts a :class:`~.QNode`, and returns a TensorFlow eager-execution-compatible QNode.

    Args:
        qnode (~pennylane.qnode.QNode): a PennyLane QNode

    Returns:
        function: the QNode as a TensorFlow function
    """
    class qnode_str(partial):
        """TensorFlow QNode"""
        # pylint: disable=too-few-public-methods

        def __str__(self):
            """String representation"""
            detail = "<QNode: device='{}', func={}, wires={}, interface=TensorFlow>"
            return detail.format(qnode.device.short_name, qnode.func.__name__, qnode.num_wires)

        def __repr__(self):
            """REPL representation"""
            return self.__str__()

        print_applied = qnode.print_applied
        jacobian = qnode.jacobian
        metric_tensor = qnode.metric_tensor

    @qnode_str
    @tf.custom_gradient
    def _TFQNode(*input_, **input_kwargs):
        # detach all input Tensors, convert to NumPy array
        args = [i.numpy() if isinstance(i, (Variable, tf.Tensor)) else i for i in input_]
        kwargs = {k:v.numpy() if isinstance(v, (Variable, tf.Tensor)) else v for k, v in input_kwargs.items()}

        # if NumPy array is scalar, convert to a Python float
        args = [i.tolist() if (isinstance(i, np.ndarray) and not i.shape) else i for i in args]
        kwargs = {k:v.tolist() if (isinstance(v, np.ndarray) and not v.shape) else v for k, v in kwargs.items()}

        # evaluate the QNode
        res = qnode(*args, **kwargs)

        if not isinstance(res, np.ndarray):
            # scalar result, cast to NumPy scalar
            res = np.array(res)

        def grad(grad_output):
            """Returns the vector-Jacobian product"""
            # evaluate the Jacobian matrix of the QNode
            jacobian = qnode.jacobian(args, **kwargs)

            grad_output_np = grad_output.numpy()

            # perform the vector-Jacobian product
            if not grad_output_np.shape:
                temp = grad_output_np * jacobian
            else:
                temp = grad_output_np.T @ jacobian

            # restore the nested structure of the input args
            grad_input = unflatten(temp.flat, args)

            if isinstance(grad_input, list):
                grad_input = [tf.convert_to_tensor(i) for i in grad_input]
            elif isinstance(grad_input, tuple):
                grad_input = tuple(tf.convert_to_tensor(i) for i in grad_input)
            else:
                grad_input = tf.convert_to_tensor(grad_input)

            return grad_input

        return res, grad

    return _TFQNode



"""
.. _pytorch_noise:

PyTorch and noisy devices
=========================

Let's revisit the original :ref:`qubit rotation <qubit_rotation>` tutorial, but instead of
using the default NumPy/autograd QNode interface, we'll use the :ref:`PyTorch interface <torch_interf>`'s
QNode.
We'll also replace the ``default.qubit`` device with a noisy ``forest.qvm`` device, to
see how the optimization responds to noisy qubits.


To follow along with this tutorial on your own computer, you will require the
following dependencies:

* The `Forest SDK <https://rigetti.com/forest>`_, which contains the quantum virtual
  machine (QVM) and quilc quantum compiler. Once installed, the QVM and quilc can be
  started by running the commands ``quilc -S`` and ``qvm -S`` in separate terminal windows.

* `PennyLane-Forest plugin <https://pennylane-forest.readthedocs.io>`_, in order
  to access the QVM as a PennyLane device. This can be installed via pip:

  .. code-block:: bash

      pip install pennylane-forest

* `PyTorch <https://pytorch.org/get-started/locally/>`_, in order to access the PyTorch
  QNode interface. Follow the link for instructions on the best way to install PyTorch
  for your system.
"""

##############################################################################
# Setting up the device
# ---------------------
#
# Once the dependencies above are installed, let's begin importing the required packages
# and setting up our quantum device.
#
# To start with, we import PennyLane, and, as we are using the PyTorch interface,
# PyTorch as well:

import pennylane as qml
import torch
from torch.autograd import Variable

##############################################################################
# Note that we do not need to import the wrapped version of NumPy provided by PennyLane,
# as we are not using the default QNode NumPy interface. If NumPy is needed, it is fine to
# import vanilla NumPy for use with PyTorch and TensorFlow.
#
# Next, we will create our device:

dev = qml.device("forest.qvm", device="2q", noisy=True)

##############################################################################
# Here, we create a noisy two-qubit system, simulated via the QVM. If we wish, we could
# also build the model on a physical device, such as the ``Aspen-1`` QPU.


##############################################################################
# Constructing the QNode
# ----------------------
#
# Now that we have initialized the device, we can construct our quantum node. Like the
# other tutorials, we use the :mod:`qnode decorator <pennylane.decorator>` to convert
# our quantum function (encoded by the circuit above) into a quantum node
# running on the QVM.


@qml.qnode(dev, interface="torch")
def circuit(phi, theta):
    qml.RX(theta, wires=0)
    qml.RZ(phi, wires=0)
    return qml.expval(qml.PauliZ(0))


##############################################################################
# To make the QNode 'PyTorch aware', we need to specify that the QNode interfaces
# with PyTorch. This is done by passing the ``interface='torch'`` keyword argument.
#
# As a result, this QNode will be set up to accept and return PyTorch tensors, and will
# also automatically calculate any analytic gradients when PyTorch performs backpropagation.


##############################################################################
# Optimization
# ------------
#
# We can now create our optimization cost function. To introduce some additional
# complexity into the system, rather than simply training the variational circuit
# to 'flip a qubit' from state :math:`\left|0\right\rangle` to state :math:`\left|1\right\rangle`, let's also
# modify the target state every 100 steps. For example, for the first 100 steps,
# the target state will be :math:`\left|1\right\rangle`; this will then change to :math:`\left|0\right\rangle`
# for steps 100 and 200, before changing back to state :math:`\left|1\right\rangle` for steps 200
# to 300, and so on.


def cost(phi, theta, step):
    target = -(-1) ** (step // 100)
    return torch.abs(circuit(phi, theta) - target) ** 2


##############################################################################
# Now that the cost function is defined, we can begin the PyTorch optimization.
# We create two variables, representing the two free parameters of the variational
# circuit, and initialize an Adam optimizer:

phi = Variable(torch.tensor(1.0), requires_grad=True)
theta = Variable(torch.tensor(0.05), requires_grad=True)
opt = torch.optim.Adam([phi, theta], lr=0.1)

##############################################################################
# As we are using the PyTorch interface, we must use PyTorch optimizers,
# *not* the built-in optimizers provided by PennyLane. The built-in optimizers
# only apply to the default NumPy/autograd interface.
#
# Optimizing the system for 400 steps:

for i in range(400):
    opt.zero_grad()
    loss = cost(phi, theta, i)
    loss.backward()
    opt.step()

##############################################################################
# We can now check the final values of the parameters, as well as the final
# circuit output and cost function:

print(phi)
print(theta)
print(circuit(phi, theta))
print(cost(phi, theta, 400))

##############################################################################
# .. rst-class:: sphx-glr-script-out
#
#  Out:
#
#  .. code-block:: none
#
#    tensor(-0.7055, requires_grad=True)
#    tensor(6.1330, requires_grad=True)
#    tensor(0.9551, dtype=torch.float64, grad_fn=<_TorchQNodeBackward>)
#    tensor(3.7162, dtype=torch.float64, grad_fn=<PowBackward0>)


##############################################################################
# As the cost function is step-dependent, this does not provide enough detail to
# determine if the optimization was successful; instead, let's plot the output
# state of the circuit over time on a Bloch sphere:
#
# .. figure:: ../../examples/figures/bloch.gif
#     :align: center
#     :target: javascript:void(0);
#
# Here, the red x is the target state of the variational circuit, and the arrow is
# the variational circuit output state. As the target state changes, the circuit
# learns to produce the new target state!


##############################################################################
# Hybrid GPU-QPU optimization
# ---------------------------
#
# As PyTorch natively supports GPU-accelerated classical processing, and Forest provides
# quantum hardware access in the form of QPUs, with very little modification, we can run
# the above code as a hybrid GPU-QPU optimization (note that to run the following
# script, you will need to be using Rigetti's QCS service):

import pennylane as qml
import torch
from torch.autograd import Variable

qpu = qml.device("forest.qpu", device="Aspen-1-2Q-B")


@qml.qnode(dev, interface="torch")
def circuit(phi, theta):
    qml.RX(theta, wires=0)
    qml.RZ(phi, wires=0)
    return qml.expval(qml.PauliZ(0))


def cost(phi, theta, step):
    target = -(-1) ** (step // 100)
    return torch.abs(circuit(phi, theta) - target) ** 2


phi = Variable(torch.tensor(1.0, device="cuda"), requires_grad=True)
theta = Variable(torch.tensor(0.05, device="cuda"), requires_grad=True)
opt = torch.optim.Adam([phi, theta], lr=0.1)

for i in range(400):
    opt.zero_grad()
    loss = cost(phi, theta, i)
    loss.backward()
    opt.step()

##############################################################################
# When using a classical interface that supports GPUs, the QNode will automatically
# copy any tensor arguments to the CPU, before applying them on the specified quantum
# device. Once done, it will return a tensor containing the QNode result, and
# automatically copy it back to the GPU for any further classical processing.
#
# .. note:: For more details on the PyTorch interface, see :ref:`torch_interf`.



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane` configuration classe :class:`Configuration`.
"""
# pylint: disable=protected-access

import contextlib
import io
import re
import sys

import pennylane as qml


def test_about():
    """
    about: Tests if the about string prints correct.
    """
    f = io.StringIO()
    with contextlib.redirect_stdout(f):
        qml.about()
    out = f.getvalue().strip()

    assert "Version:" in out
    pl_version_match = re.search(r"Version:\s+([\S]+)\n", out).group(1)
    assert qml.version().replace("-", ".") in pl_version_match
    assert "Numpy version" in out
    assert "Scipy version" in out
    assert "default.qubit" in out
    assert "default.gaussian" in out



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane.plugin.DefaultGaussian` device.
"""
# pylint: disable=protected-access,cell-var-from-loop,no-self-use

import pytest
from scipy.special import factorial as fac
from scipy.linalg import block_diag
import numpy as np
import numpy.random

import pennylane as qml
from pennylane.plugins.default_gaussian import (
    fock_prob,
    rotation, squeezing, quadratic_phase, beamsplitter, two_mode_squeezing, controlled_addition, controlled_phase,
    vacuum_state, coherent_state, squeezed_state, displaced_squeezed_state, thermal_state,
    DefaultGaussian)



U = np.array(
    [[0.83645892-0.40533293j, -0.20215326+0.30850569j],
     [-0.23889780-0.28101519j, -0.88031770-0.29832709j]]
)


U2 = np.array(
    [[-0.07843244-3.57825948e-01j, 0.71447295-5.38069384e-02j, 0.20949966+6.59100734e-05j, -0.50297381+2.35731613e-01j],
     [-0.26626692+4.53837083e-01j, 0.27771991-2.40717436e-01j, 0.41228017-1.30198687e-01j, 0.01384490-6.33200028e-01j],
     [-0.69254712-2.56963068e-02j, -0.15484858+6.57298384e-02j, -0.53082141+7.18073414e-02j, -0.41060450-1.89462315e-01j],
     [-0.09686189-3.15085273e-01j, -0.53241387-1.99491763e-01j, 0.56928622+3.97704398e-01j, -0.28671074-6.01574497e-02j]]
)


H = np.array(
    [[1.02789352, 1.61296440-0.3498192j],
     [1.61296440+0.3498192j, 1.23920938+0j]]
)


hbar = 2

def prep_par(par, op):
    "Convert par into a list of parameters that op expects."
    if op.par_domain == 'A':
        return [np.diag([x, 1]) for x in par]
    return par


@pytest.fixture(scope="function")
def gaussian_device_1_wire():
    """Fixture of a default.gaussian device with 1 wire."""
    return qml.device('default.gaussian', wires=1)

@pytest.fixture(scope="function")
def gaussian_device_2_wires():
    """Fixture of a default.gaussian device with 2 wires."""
    return qml.device('default.gaussian', wires=2)

@pytest.fixture(scope="function")
def gaussian_device_3_wires():
    """Fixture of a default.gaussian device with 3 wires."""
    return qml.device('default.gaussian', wires=3)

gaussian_dev = gaussian_device_2_wires  # alias



class TestExceptions:
    """Tests that default.gaussian throws the correct error messages"""

    def test_sample_exception(self):
        """Test that default.gaussian raises an exception if sampling is attempted."""

        dev = qml.device('default.gaussian', wires=1)

        @qml.qnode(dev)
        def circuit():
            return qml.sample(qml.NumberOperator(0))
            raise NotImplementedError()

        with pytest.raises(NotImplementedError, match="default.gaussian does not support sampling NumberOperator"):
            circuit()

class TestAuxillaryFunctions:
    """Tests the auxillary functions"""

    def test_fock_prob(self, tol):
        """Test fock_prob returns the correct Fock probabilities"""

        # an arbitrary two-mode Gaussian state generated using Strawberry Fields
        mu = np.array([0.6862, 0.4002, 0.09, 0.558]) * np.sqrt(hbar)
        cov = np.array(
            [[0.50750512, -0.04125979, -0.21058229, -0.07866912],
             [-0.04125979, 0.50750512, -0.07866912, -0.21058229],
             [-0.21058229, -0.07866912, 0.95906208, 0.27133391],
             [-0.07866912, -0.21058229, 0.27133391, 0.95906208]]
        ) * hbar

        # expected Fock state probabilities
        events = [(0, 0), (0, 1), (1, 1), (2, 3)]
        probs = [0.430461524043, 0.163699407559, 0.0582788388927, 0.00167706931355]

        for idx, e in enumerate(events):
            res = fock_prob(mu, cov, e, hbar=hbar)
            assert res == pytest.approx(probs[idx], abs=tol)


class TestGates:
    """Gate tests."""

    def test_rotation(self, tol):
        """Test the Fourier transform of a displaced state."""
        # pylint: disable=invalid-unary-operand-type

        alpha = 0.23+0.12j
        S = rotation(np.pi/2)

        # apply to a coherent state. F{x, p} -> {-p, x}
        out = S @ np.array([alpha.real, alpha.imag])*np.sqrt(2*hbar)
        expected = np.array([-alpha.imag, alpha.real])*np.sqrt(2*hbar)
        assert out == pytest.approx(expected, abs=tol)

    def test_squeezing(self, tol):
        """Test the squeezing symplectic transform."""

        r = 0.543
        phi = 0.123
        S = squeezing(r, phi)

        # apply to an identity covariance matrix
        out = S @ S.T
        expected = rotation(phi/2) @ np.diag(np.exp([-2*r, 2*r])) @ rotation(phi/2).T
        assert out == pytest.approx(expected, abs=tol)

    def test_quadratic_phase(self, tol):
        """Test the quadratic phase symplectic transform."""

        s = 0.543
        S = quadratic_phase(s)

        # apply to a coherent state. P[x, p] -> [x, p+sx]
        alpha = 0.23+0.12j
        out = S @ np.array([alpha.real, alpha.imag])*np.sqrt(2*hbar)
        expected = np.array([alpha.real, alpha.imag+s*alpha.real])*np.sqrt(2*hbar)
        assert out == pytest.approx(expected, abs=tol)

    def test_beamsplitter(self, tol):
        """Test the beamsplitter symplectic transform."""

        theta = 0.543
        phi = 0.312
        S = beamsplitter(theta, phi)

        # apply to a coherent state. BS|a1, a2> -> |ta1-r^*a2, ra1+ta2>
        a1 = 0.23+0.12j
        a2 = 0.23+0.12j
        out = S @ np.array([a1.real, a2.real, a1.imag, a2.imag])*np.sqrt(2*hbar)

        T = np.cos(theta)
        R = np.exp(1j*phi)*np.sin(theta)
        a1out = T*a1 - R.conj()*a2
        a2out = R*a2 + T*a1
        expected = np.array([a1out.real, a2out.real, a1out.imag, a2out.imag])*np.sqrt(2*hbar)
        assert out == pytest.approx(expected, abs=tol)

    def test_two_mode_squeezing(self, tol):
        """Test the two mode squeezing symplectic transform."""

        r = 0.543
        phi = 0.123
        S = two_mode_squeezing(r, phi)

        # test that S = B^\dagger(pi/4, 0) [S(z) x S(-z)] B(pi/4)
        B = beamsplitter(np.pi/4, 0)
        Sz = block_diag(squeezing(r, phi), squeezing(-r, phi))[:, [0, 2, 1, 3]][[0, 2, 1, 3]]
        expected = B.conj().T @ Sz @ B
        assert S == pytest.approx(expected, abs=tol)

        # test that S |a1, a2> = |ta1+ra2, ta2+ra1>
        a1 = 0.23+0.12j
        a2 = 0.23+0.12j
        out = S @ np.array([a1.real, a2.real, a1.imag, a2.imag])*np.sqrt(2*hbar)

        T = np.cosh(r)
        R = np.exp(1j*phi)*np.sinh(r)
        a1out = T*a1 + R*np.conj(a2)
        a2out = T*a2 + R*np.conj(a1)
        expected = np.array([a1out.real, a2out.real, a1out.imag, a2out.imag])*np.sqrt(2*hbar)
        assert out == pytest.approx(expected, abs=tol)

    def test_controlled_addition(self, tol):
        """Test the CX symplectic transform."""

        s = 0.543
        S = controlled_addition(s)

        # test that S = B(theta+pi/2, 0) [S(z) x S(-z)] B(theta, 0)
        r = np.arcsinh(-s/2)
        theta = 0.5*np.arctan2(-1/np.cosh(r), -np.tanh(r))
        Sz = block_diag(squeezing(r, 0), squeezing(-r, 0))[:, [0, 2, 1, 3]][[0, 2, 1, 3]]

        expected = beamsplitter(theta+np.pi/2, 0) @ Sz @ beamsplitter(theta, 0)
        assert S == pytest.approx(expected, abs=tol)

        # test that S[x1, x2, p1, p2] -> [x1, x2+sx1, p1-sp2, p2]
        x1 = 0.5432
        x2 = -0.453
        p1 = 0.154
        p2 = -0.123
        out = S @ np.array([x1, x2, p1, p2])*np.sqrt(2*hbar)
        expected = np.array([x1, x2+s*x1, p1-s*p2, p2])*np.sqrt(2*hbar)
        assert out == pytest.approx(expected, abs=tol)

    def test_controlled_phase(self, tol):
        """Test the CZ symplectic transform."""

        s = 0.543
        S = controlled_phase(s)

        # test that S = R_2(pi/2) CX(s) R_2(pi/2)^\dagger
        R2 = block_diag(np.identity(2), rotation(np.pi/2))[:, [0, 2, 1, 3]][[0, 2, 1, 3]]
        expected = R2 @ controlled_addition(s) @ R2.conj().T
        assert S == pytest.approx(expected, abs=tol)

        # test that S[x1, x2, p1, p2] -> [x1, x2, p1+sx2, p2+sx1]
        x1 = 0.5432
        x2 = -0.453
        p1 = 0.154
        p2 = -0.123
        out = S @ np.array([x1, x2, p1, p2])*np.sqrt(2*hbar)
        expected = np.array([x1, x2, p1+s*x2, p2+s*x1])*np.sqrt(2*hbar)
        assert out == pytest.approx(expected, abs=tol)


class TestStates:
    """State tests."""

    def test_vacuum_state(self, tol):
        """Test the vacuum state is correct."""
        wires = 3
        means, cov = vacuum_state(wires, hbar=hbar)
        assert means == pytest.approx(np.zeros([2*wires]), abs=tol)
        assert cov == pytest.approx(np.identity(2*wires)*hbar/2, abs=tol)

    def test_coherent_state(self, tol):
        """Test the coherent state is correct."""
        a = 0.432-0.123j
        means, cov = coherent_state(a, hbar=hbar)
        assert means == pytest.approx(np.array([a.real, a.imag])*np.sqrt(2*hbar), abs=tol)
        assert cov == pytest.approx(np.identity(2)*hbar/2, abs=tol)

    def test_squeezed_state(self, tol):
        """Test the squeezed state is correct."""
        r = 0.432
        phi = 0.123
        means, cov = squeezed_state(r, phi, hbar=hbar)

        # test vector of means is zero
        assert means == pytest.approx(np.zeros([2]), abs=tol)

        R = rotation(phi/2)
        expected = R @ np.array([[np.exp(-2*r), 0],
                                 [0, np.exp(2*r)]]) * hbar/2 @ R.T
        # test covariance matrix is correct
        assert cov == pytest.approx(expected, abs=tol)

    def test_displaced_squeezed_state(self, tol):
        """Test the displaced squeezed state is correct."""
        alpha = 0.541+0.109j
        a = abs(alpha)
        phi_a = np.angle(alpha)
        r = 0.432
        phi_r = 0.123
        means, cov = displaced_squeezed_state(a, phi_a, r, phi_r, hbar=hbar)

        # test vector of means is correct
        assert means == pytest.approx(np.array([alpha.real, alpha.imag])*np.sqrt(2*hbar), abs=tol)

        R = rotation(phi_r/2)
        expected = R @ np.array([[np.exp(-2*r), 0],
                                 [0, np.exp(2*r)]]) * hbar/2 @ R.T
        # test covariance matrix is correct
        assert cov == pytest.approx(expected, abs=tol)

    def thermal_state(self, tol):
        """Test the thermal state is correct."""
        nbar = 0.5342
        means, cov = thermal_state(nbar, hbar=hbar)
        assert means == pytest.approx(np.zeros([2]), abs=tol)
        assert np.all((cov.diag*2/hbar-1)/2 == nbar)


class TestDefaultGaussianDevice:
    """Test the default gaussian device. The test ensures that the device is properly
    applying gaussian operations and calculating the correct observables."""

    def test_operation_map(self, gaussian_dev):
        """Test that default Gaussian device supports all PennyLane Gaussian CV gates."""

        non_supported = {'FockDensityMatrix',
                         'FockStateVector',
                         'FockState',
                         'CrossKerr',
                         'CatState',
                         'CubicPhase',
                         'Kerr'}

        assert set(qml.ops._cv__ops__) - non_supported == set(gaussian_dev._operation_map)

    def test_observable_map(self, gaussian_dev):
        """Test that default Gaussian device supports all PennyLane Gaussian continuous observables."""
        assert set(qml.ops._cv__obs__)|{'Identity'}-{'Heterodyne'} == set(gaussian_dev._observable_map)

    def test_apply(self, gaussian_dev, tol):
        """Test the application of gates to a state"""

        # loop through all supported operations
        for gate_name, fn in gaussian_dev._operation_map.items():
            #log.debug("\tTesting %s gate...", gate_name)
            gaussian_dev.reset()

            # start in the displaced squeezed state
            alpha = 0.542+0.123j
            a = abs(alpha)
            phi_a = np.angle(alpha)
            r = 0.652
            phi_r = -0.124

            gaussian_dev.apply('DisplacedSqueezedState', wires=[0], par=[a, phi_a, r, phi_r])
            gaussian_dev.apply('DisplacedSqueezedState', wires=[1], par=[a, phi_a, r, phi_r])

            # get the equivalent pennylane operation class
            op = qml.ops.__getattribute__(gate_name)
            # the list of wires to apply the operation to
            w = list(range(op.num_wires))

            if op.par_domain == 'A':
                # the parameter is an array
                if gate_name == 'GaussianState':
                    p = [np.array([0.432, 0.123, 0.342, 0.123]), np.diag([0.5234]*4)]
                    w = list(range(2))
                    expected_out = p
                elif gate_name == 'Interferometer':
                    w = list(range(2))
                    p = [U]
                    S = fn(*p)
                    expected_out = S @ gaussian_dev._state[0], S @ gaussian_dev._state[1] @ S.T
            else:
                # the parameter is a float
                p = [0.432423, -0.12312, 0.324, 0.751][:op.num_params]

                if gate_name == 'Displacement':
                    alpha = p[0]*np.exp(1j*p[1])
                    state = gaussian_dev._state
                    mu = state[0].copy()
                    mu[w[0]] += alpha.real*np.sqrt(2*hbar)
                    mu[w[0]+2] += alpha.imag*np.sqrt(2*hbar)
                    expected_out = mu, state[1]
                elif 'State' in gate_name:
                    mu, cov = fn(*p, hbar=hbar)
                    expected_out = gaussian_dev._state
                    expected_out[0][[w[0], w[0]+2]] = mu

                    ind = np.concatenate([np.array([w[0]]), np.array([w[0]])+2])
                    rows = ind.reshape(-1, 1)
                    cols = ind.reshape(1, -1)
                    expected_out[1][rows, cols] = cov
                else:
                    # if the default.gaussian is an operation accepting parameters,
                    # initialise it using the parameters generated above.
                    S = fn(*p)

                    # calculate the expected output
                    if op.num_wires == 1:
                        # reorder from symmetric ordering to xp-ordering
                        S = block_diag(S, np.identity(2))[:, [0, 2, 1, 3]][[0, 2, 1, 3]]

                    expected_out = S @ gaussian_dev._state[0], S @ gaussian_dev._state[1] @ S.T

            gaussian_dev.apply(gate_name, wires=w, par=p)

            # verify the device is now in the expected state
            assert gaussian_dev._state[0] == pytest.approx(expected_out[0], abs=tol)
            assert gaussian_dev._state[1] == pytest.approx(expected_out[1], abs=tol)

    def test_apply_errors(self, gaussian_dev):
        """Test that apply fails for incorrect state preparation"""

        with pytest.raises(ValueError, match='incorrect size for the number of subsystems'):
            p = [thermal_state(0.5)]
            gaussian_dev.apply('GaussianState', wires=[0], par=[p])

        with pytest.raises(ValueError, match='Incorrect number of subsystems'):
            p = U
            gaussian_dev.apply('Interferometer', wires=[0], par=[p])

        with pytest.raises(ValueError, match="Invalid target subsystems provided in 'wires' argument"):
            p = U2
            #dev = DefaultGaussian(wires=4, shots=1000, hbar=hbar)
            gaussian_dev.apply('Interferometer', wires=[0, 1, 2], par=[p])

    def test_expectation(self, tol):
        """Test that expectation values are calculated correctly"""

        dev = qml.device('default.gaussian', wires=1, hbar=hbar)

        # test correct mean for <n> of a displaced thermal state
        nbar = 0.5431
        alpha = 0.324-0.59j
        dev.apply('ThermalState', wires=[0], par=[nbar])
        dev.apply('Displacement', wires=[0], par=[alpha, 0])
        mean = dev.expval('NumberOperator', [0], [])
        assert mean == pytest.approx(np.abs(alpha)**2+nbar, abs=tol)

        # test correct mean for Homodyne P measurement
        alpha = 0.324-0.59j
        dev.apply('CoherentState', wires=[0], par=[alpha])
        mean = dev.expval('P', [0], [])
        assert mean == pytest.approx(alpha.imag*np.sqrt(2*hbar), abs=tol)

        # test correct mean for Homodyne measurement
        mean = dev.expval('QuadOperator', [0], [np.pi/2])
        assert mean == pytest.approx(alpha.imag*np.sqrt(2*hbar), abs=tol)

        # test correct mean for number state expectation |<n|alpha>|^2
        # on a coherent state
        for n in range(3):
            mean = dev.expval('FockStateProjector', [0], [np.array([n])])
            expected = np.abs(np.exp(-np.abs(alpha)**2/2)*alpha**n/np.sqrt(fac(n)))**2
            assert mean == pytest.approx(expected, abs=tol)

        # test correct mean for number state expectation |<n|S(r)>|^2
        # on a squeezed state
        n = 1
        r = 0.4523
        dev.apply('SqueezedState', wires=[0], par=[r, 0])
        mean = dev.expval('FockStateProjector', [0], [np.array([2*n])])
        expected = np.abs(np.sqrt(fac(2*n))/(2**n*fac(n))*(-np.tanh(r))**n/np.sqrt(np.cosh(r)))**2
        assert mean == pytest.approx(expected, abs=tol)

    def test_variance_displaced_thermal_mean_photon(self, tol):
        """test correct variance for <n> of a displaced thermal state"""
        dev = qml.device('default.gaussian', wires=1, hbar=hbar)

        nbar = 0.5431
        alpha = 0.324-0.59j
        dev.apply('ThermalState', wires=[0], par=[nbar])
        dev.apply('Displacement', wires=[0], par=[alpha, 0])
        var = dev.var('NumberOperator', [0], [])
        assert var == pytest.approx(nbar**2+nbar+np.abs(alpha)**2*(1+2*nbar), abs=tol)

    def test_variance_coherent_homodyne(self, tol):
        """test correct variance for Homodyne P measurement"""
        dev = qml.device('default.gaussian', wires=1, hbar=hbar)

        alpha = 0.324-0.59j
        dev.apply('CoherentState', wires=[0], par=[alpha])
        var = dev.var('P', [0], [])
        assert var == pytest.approx(hbar/2, abs=tol)

        # test correct mean and variance for Homodyne measurement
        var = dev.var('QuadOperator', [0], [np.pi/2])
        assert var == pytest.approx(hbar/2, abs=tol)

    def test_variance_coherent_numberstate(self, tol):
        """test correct variance for number state expectation |<n|alpha>|^2
        on a coherent state
        """
        dev = qml.device('default.gaussian', wires=1, hbar=hbar)

        alpha = 0.324-0.59j

        dev.apply('CoherentState', wires=[0], par=[alpha])

        for n in range(3):
            var = dev.var('FockStateProjector', [0], [np.array([n])])
            mean = np.abs(np.exp(-np.abs(alpha)**2/2)*alpha**n/np.sqrt(fac(n)))**2
            assert var == pytest.approx(mean*(1-mean), abs=tol)

    def test_variance_squeezed_numberstate(self, tol):
        """test correct variance for number state expectation |<n|S(r)>|^2
        on a squeezed state
        """
        dev = qml.device('default.gaussian', wires=1, hbar=hbar)

        n = 1
        r = 0.4523
        dev.apply('SqueezedState', wires=[0], par=[r, 0])
        var = dev.var('FockStateProjector', [0], [np.array([2*n])])
        mean = np.abs(np.sqrt(fac(2*n))/(2**n*fac(n))*(-np.tanh(r))**n/np.sqrt(np.cosh(r)))**2
        assert var == pytest.approx(mean*(1-mean), abs=tol)

    def test_reduced_state(self, gaussian_dev, tol):
        """Test reduced state"""

        # Test error is raised if requesting a non-existant subsystem
        with pytest.raises(ValueError, match="specified wires cannot be larger than the number of subsystems"):
            gaussian_dev.reduced_state([6, 4])

        # Test requesting via an integer
        res = gaussian_dev.reduced_state(0)
        expected = gaussian_dev.reduced_state([0])
        assert res[0] == pytest.approx(expected[0], abs=tol)
        assert res[1] == pytest.approx(expected[1], abs=tol)

        # Test requesting all wires returns the full state
        res = gaussian_dev.reduced_state([0, 1])
        expected = gaussian_dev._state
        assert res[0] == pytest.approx(expected[0], abs=tol)
        assert res[1] == pytest.approx(expected[1], abs=tol)



def input_logger(*args):
    """Helper function for monkeypatch: logs its input."""
    input_logger.args = args
    return np.array([1, 2, 3, 4, 5])


class TestSample:
    """Tests that sampling is correctly implemented."""

    @pytest.mark.parametrize("alpha", [0.324-0.59j, 2.3+1.2j, 1.3j, -1.2])
    def test_sampling_parameters_coherent(self, tol, gaussian_device_1_wire, alpha, monkeypatch):
        """Tests that the np.random.normal is called with the correct parameters that reflect
           the underlying distribution for a coherent state."""

        mean = alpha.imag*np.sqrt(2*gaussian_device_1_wire.hbar)
        std = gaussian_device_1_wire.hbar/2
        gaussian_device_1_wire.apply('CoherentState', wires=[0], par=[alpha])

        with monkeypatch.context() as m:
            m.setattr(numpy.random, 'normal', input_logger)
            sample = gaussian_device_1_wire.sample('P', [0], [])
            assert np.allclose(input_logger.args, [mean, std, gaussian_device_1_wire.shots], atol=tol, rtol=0)

    @pytest.mark.parametrize("alpha", [0.324-0.59j, 2.3+1.2j, 1.3j, -1.2])
    def test_sampling_parameters_coherent_quad_operator(self, tol, gaussian_device_1_wire, alpha, monkeypatch):
        """Tests that the np.random.normal is called with the correct parameters that reflect
           the underlying distribution for a coherent state when using QuadOperator."""

        mean = alpha.imag*np.sqrt(2*gaussian_device_1_wire.hbar)
        std = gaussian_device_1_wire.hbar/2
        gaussian_device_1_wire.apply('CoherentState', wires=[0], par=[alpha])

        with monkeypatch.context() as m:
            m.setattr(numpy.random, 'normal', input_logger)
            sample = gaussian_device_1_wire.sample('QuadOperator', [0], [np.pi/2])
            assert np.allclose(input_logger.args, [mean, std, gaussian_device_1_wire.shots], atol=tol, rtol=0)

    @pytest.mark.parametrize("r,phi", [(1.0, 0.0)])
    def test_sampling_parameters_squeezed(self, tol, gaussian_device_1_wire, r, phi, monkeypatch):
        """Tests that the np.random.normal is called with the correct parameters that reflect
           the underlying distribution for a squeezed state."""

        mean = 0.0
        std = np.sqrt(gaussian_device_1_wire.hbar*np.exp(2*r)/2)
        gaussian_device_1_wire.apply('SqueezedState', wires=[0], par=[r, phi])

        with monkeypatch.context() as m:
            m.setattr(numpy.random, 'normal', input_logger)
            sample = gaussian_device_1_wire.sample('P', [0], [])
            assert np.allclose(input_logger.args, [mean, std, gaussian_device_1_wire.shots], atol=tol, rtol=0)

    @pytest.mark.parametrize("observable,n_sample", [('P', 10), ('P', 25), ('X', 1), ('X', 16)])
    def test_sample_shape_and_dtype(self, gaussian_device_2_wires, observable, n_sample):
        """Test that the sample function outputs samples of the right size"""

        gaussian_device_2_wires.shots = n_sample
        sample = gaussian_device_2_wires.sample(observable, [0], [])

        assert np.array_equal(sample.shape, (n_sample,))
        assert sample.dtype == np.dtype("float")

    def test_sample_error_multi_wire(self, gaussian_device_2_wires):
        """Test that the sample function raises an error if multiple wires are given"""

        with pytest.raises(ValueError, match="Only one mode can be measured in homodyne"):
            sample = gaussian_device_2_wires.sample('P', [0, 1], [])

    @pytest.mark.parametrize("observable", set(qml.ops.cv.obs) - set(['P', 'X', 'QuadOperator']))
    def test_sample_error_unsupported_observable(self, gaussian_device_2_wires, observable):
        """Test that the sample function raises an error if the given observable is not supported"""

        with pytest.raises(NotImplementedError, match="default.gaussian does not support sampling"):
            sample = gaussian_device_2_wires.sample(observable, [0], [])


class TestDefaultGaussianIntegration:
    """Integration tests for default.gaussian. This test ensures it integrates
    properly with the PennyLane interface, in particular QNode."""

    def test_load_default_gaussian_device(self):
        """Test that the default plugin loads correctly"""

        dev = qml.device('default.gaussian', wires=2, hbar=2)
        assert dev.num_wires == 2
        assert dev.shots == 1000
        assert dev.analytic == True
        assert dev.hbar == 2
        assert dev.short_name == 'default.gaussian'

    def test_args(self):
        """Test that the plugin requires correct arguments"""

        with pytest.raises(TypeError, match="missing 1 required positional argument: 'wires'"):
            qml.device('default.gaussian')

    @pytest.mark.parametrize("g", set(qml.ops.__all_ops__) - set(DefaultGaussian._operation_map.keys()))
    def test_unsupported_gates(self, g, gaussian_device_3_wires):
        """Test error is raised with unsupported gates"""

        op = getattr(qml.ops, g)
        if op.num_wires <= 0:
            wires = list(range(3))
        else:
            wires = list(range(op.num_wires))

        @qml.qnode(gaussian_device_3_wires)
        def circuit(*x):
            """Test quantum function"""
            x = prep_par(x, op)
            op(*x, wires=wires)
            return qml.expval(qml.X(0)) if issubclass(op, qml.operation.CV) else qml.expval(qml.PauliZ(0))

        x = np.random.random([op.num_params])
        with pytest.raises(qml.DeviceError, match="Gate {} not supported on device default.gaussian".format(g)):
            circuit(*x)

    @pytest.mark.parametrize("g", set(qml.ops.__all_obs__) - set(DefaultGaussian._observable_map.keys()))
    def test_unsupported_observables(self, g, gaussian_dev):
        """Test error is raised with unsupported observables."""

        op = getattr(qml.ops, g)
        if op.num_wires <= 0:
            wires = list(range(2))
        else:
            wires = list(range(op.num_wires))

        @qml.qnode(gaussian_dev)
        def circuit(*x):
            """Test quantum function"""
            x = prep_par(x, op)
            return qml.expval(op(*x, wires=wires))

        x = np.random.random([op.num_params])
        with pytest.raises(qml.DeviceError, match="Observable {} not supported on device default.gaussian".format(g)):
            circuit(*x)

    def test_gaussian_circuit(self, tol):
        """Test that the default gaussian plugin provides correct result for simple circuit"""
        dev = qml.device('default.gaussian', wires=1)

        p = 0.543

        @qml.qnode(dev)
        def circuit(x):
            """Test quantum function"""
            qml.Displacement(x, 0, wires=0)
            return qml.expval(qml.X(0))

        assert circuit(p) == pytest.approx(p*np.sqrt(2*hbar), abs=tol)

    def test_gaussian_identity(self, tol):
        """Test that the default gaussian plugin provides correct result for the identity expectation"""
        dev = qml.device('default.gaussian', wires=1)

        p = 0.543

        @qml.qnode(dev)
        def circuit(x):
            """Test quantum function"""
            qml.Displacement(x, 0, wires=0)
            return qml.expval(qml.Identity(0))

        assert circuit(p) == pytest.approx(1, abs=tol)

    def test_nonzero_shots(self, tol):
        """Test that the default gaussian plugin provides correct result for high shot number"""

        shots = 10**4
        dev = qml.device('default.gaussian', wires=1, shots=shots)

        p = 0.543

        @qml.qnode(dev)
        def circuit(x):
            """Test quantum function"""
            qml.Displacement(x, 0, wires=0)
            return qml.expval(qml.X(0))

        runs = []
        for _ in range(100):
            runs.append(circuit(p))

        assert np.mean(runs) == pytest.approx(p*np.sqrt(2*hbar), abs=tol)

    @pytest.mark.parametrize("g, qop", set(DefaultGaussian._operation_map.items()))
    def test_supported_gates(self, g, qop, gaussian_dev):
        """Test that all supported gates work correctly"""
        a = 0.312
        dev = gaussian_dev
        dev.reset()

        assert dev.supports_operation(g)

        op = getattr(qml.ops, g)
        if op.num_wires <= 0:
            wires = list(range(2))
        else:
            wires = list(range(op.num_wires))

        @qml.qnode(gaussian_dev)
        def circuit(*x):
            """Reference quantum function"""
            qml.Displacement(a, 0, wires=[0])
            op(*x, wires=wires)
            return qml.expval(qml.X(0))

        # compare to reference result
        def reference(*x):
            """reference circuit"""
            if g == 'GaussianState':
                return x[0][0]

            if g == 'Displacement':
                alpha = x[0]*np.exp(1j*x[1])
                return (alpha+a).real*np.sqrt(2*hbar)

            if 'State' in g:
                mu, _ = qop(*x, hbar=hbar)
                return mu[0]

            S = qop(*x)

            # calculate the expected output
            if op.num_wires == 1:
                S = block_diag(S, np.identity(2))[:, [0, 2, 1, 3]][[0, 2, 1, 3]]

            return (S @ np.array([a.real, a.imag, 0, 0])*np.sqrt(2*hbar))[0]

        if g == 'GaussianState':
            p = [np.array([0.432, 0.123, 0.342, 0.123]), np.diag([0.5234]*4)]
        elif g == 'Interferometer':
            p = [np.array(U)]
        else:
            p = [0.432423, -0.12312, 0.324, 0.763][:op.num_params]

        assert circuit(*p) == reference(*p)



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains the CircuitGraph class which is used to generate a DAG (directed acyclic graph)
based on operation and observable queues.
"""
from collections import namedtuple
import networkx as nx

from .utils import _flatten


def _by_idx(x):
    """Sorting key for Operations: queue index aka temporal order."""
    return x.queue_idx


Layer = namedtuple("Layer", ["ops", "param_inds"])
"""Parametrized layer of the circuit.

Args:

    ops (list[Operation]): parametrized operations in the layer
    param_inds (tuple[int]): corresponding free parameter indices
"""
#TODO define what a layer is

LayerData = namedtuple("LayerData", ["pre_ops", "ops", "param_inds", "post_ops"])
"""Parametrized layer of the circuit.

Args:
    pre_ops (list[Operation]): operations that precede the layer
    ops (list[Operation]): parametrized operations in the layer
    param_inds (tuple[int]): corresponding free parameter indices
    post_ops (list[Operation]): operations that succeed the layer
"""

class CircuitGraph:
    """Represents a quantum circuit as a directed acyclic graph.

    Args:
        ops (Iterable[Operation]): quantum operations constituting the circuit, in temporal order
        variable_deps (dict[int, list[ParDep]]): Free parameters of the quantum circuit.
            The dictionary key is the parameter index.
            The first element of the ParDep namedtuple is the operation index,
            the second the index of the parameter within the operation.
    """
    def __init__(self, ops, variable_deps):
        self.variable_deps = variable_deps

        self._grid = {}
        """dict[int, list[Operation]]: dictionary representing the quantum circuit as a grid.
        Here, the key is the wire number, and the value is a list containing the operations on that wire.
        """
        for k, op in enumerate(ops):
            op.queue_idx = k  # store the queue index in the Operation
            for w in set(_flatten(op.wires)):  # flatten the nested wires lists of Tensor observables
                # Add op to the grid, to the end of wire w
                self._grid.setdefault(w, []).append(op)

        self._graph = nx.DiGraph()  #: nx.DiGraph: DAG representation of the quantum circuit
        # Iterate over each (populated) wire in the grid
        for wire in self._grid.values():
            # Add the first operation on the wire to the graph
            # This operation does not depend on any others
            self._graph.add_node(wire[0])

            for i in range(1, len(wire)):
                # For subsequent operations on the wire:
                if wire[i] not in self._graph:
                    # Add them to the graph if they are not already
                    # in the graph (multi-qubit operations might already have been placed)
                    self._graph.add_node(wire[i])

                # Create an edge between this operation and the previous operation
                self._graph.add_edge(wire[i - 1], wire[i])

    @property
    def observables(self):
        """Observables in the circuit, sorted by queue index.

        Returns:
            list[Observable]: observables
        """
        nodes = [node for node in self._graph.nodes if getattr(node, 'return_type', None) is not None]
        return sorted(nodes, key=_by_idx)

    @property
    def operations(self):
        """Non-observable Operations in the circuit, sorted by queue index.

        The complement of :meth:`QNode.observables`. Together they return every :class:`Operation`
        instance in the circuit.

        Returns:
            list[Operation]: operations
        """
        nodes = [node for node in self._graph.nodes if getattr(node, 'return_type', None) is None]
        return sorted(nodes, key=_by_idx)

    @property
    def graph(self):
        """The graph representation of the quantum circuit.

        The graph has nodes representing :class:`.Operation` instances,
        and directed edges pointing from nodes to their immediate dependents/successors.

        Returns:
            networkx.DiGraph: the directed acyclic graph representing the quantum circuit
        """
        return self._graph

    def wire_indices(self, wire):
        """Operation indices on the given wire.

        Args:
            wire (int): wire to examine

        Returns:
            list[int]: indices of operations on the wire, in temporal order
        """
        return [op.queue_idx for op in self._grid[wire]]

    def ancestors(self, ops):
        """Ancestor operations of a given set of operations.

        Args:
            ops (Iterable[Operation]): set of operations in the circuit

        Returns:
            set[Operation]: ancestors of the given operations
        """
        return set().union(*(nx.dag.ancestors(self._graph, o) for o in ops)) - set(ops)

    def descendants(self, ops):
        """Descendant operations of a given set of operations.

        Args:
            ops (Iterable[Operation]): set of operations in the circuit

        Returns:
            set[Operation]: descendants of the given operations
        """
        return set().union(*(nx.dag.descendants(self._graph, o) for o in ops)) - set(ops)

    def _in_topological_order(self, ops):
        """Sorts a set of operations in the circuit in a topological order.

        Args:
            ops (Iterable[Operation]): set of operations in the circuit

        Returns:
            Iterable[Operation]: same set of operations, topologically ordered
        """
        G = nx.DiGraph(self._graph.subgraph(ops))
        return nx.dag.topological_sort(G)

    def ancestors_in_order(self, ops):
        """Operation ancestors in a topological order.

        Currently the topological order is determined by the queue index.

        Args:
            ops (Iterable[Operation]): set of operations in the circuit

        Returns:
            list[Operation]: ancestors of the given operations, topologically ordered
        """
        #return self._in_topological_order(self.ancestors(ops))  # an abitrary topological order
        return sorted(self.ancestors(ops), key=_by_idx)

    def descendants_in_order(self, ops):
        """Operation descendants in a topological order.

        Currently the topological order is determined by the queue index.

        Args:
            ops (Iterable[Operation]): set of operations in the circuit

        Returns:
            list[Operation]: descendants of the given operations, topologically ordered
        """
        return sorted(self.descendants(ops), key=_by_idx)

    @property
    def layers(self):
        """Identify the parametrized layer structure of the circuit.

        Returns:
            list[Layer]: layers of the circuit
        """
        # FIXME maybe layering should be greedier, for example [a0 b0 c1 d1] should layer as [a0 c1], [b0, d1] and not [a0], [b0 c1], [d1]
        # keep track of the current layer
        current = Layer([], [])
        layers = [current]

        # sort vars by first occurrence of the var in the ops queue
        variable_ops_sorted = sorted(self.variable_deps.items(), key=lambda x: x[1][0].op.queue_idx)

        # iterate over all parameters
        for param_idx, gate_param_tuple in variable_ops_sorted:
            # iterate over ops depending on that param
            for op, _ in gate_param_tuple:
                # get all predecessor ops of the op
                sub = self.ancestors((op,))

                # check if any of the dependents are in the
                # currently assembled layer
                if set(current.ops) & sub:
                    # operation depends on current layer, start a new layer
                    current = Layer([], [])
                    layers.append(current)

                # store the parameters and ops indices for the layer
                current.ops.append(op)
                current.param_inds.append(param_idx)

        return layers

    def iterate_layers(self):
        """Parametrized layers of the circuit.

        Returns:
            Iterable[LayerData]: layers with extra metadata
        """
        # iterate through each layer
        for ops, param_inds in self.layers:
            pre_queue = self.ancestors_in_order(ops)
            post_queue = self.descendants_in_order(ops)
            yield LayerData(pre_queue, ops, tuple(param_inds), post_queue)

    def update_node(self, old, new):
        """Replaces the given circuit graph node with a new one.

        Args:
            old (Operation): node to replace
            new (Operation): replacement
        """
        # TODO Does alter the graph edges in any way. variable_deps is not changed, _grid is not changed. Dangerous, do we need this?
        if new.wires != old.wires:
            raise ValueError('The new Operation must act on the same wires as the old one.')
        nx.relabel_nodes(self._graph, {old: new}, copy=False)  # change the graph in place
        new.queue_idx = old.queue_idx



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane.plugin.DefaultGaussian` device.
"""
# pylint: disable=protected-access,cell-var-from-loop

from scipy.linalg import block_diag
import pytest

import pennylane
from pennylane import numpy as np
from pennylane.ops import cv


s_vals = np.linspace(-3, 3, 13)
phis = np.linspace(-2 * np.pi, 2 * np.pi, 11)
mags = np.linspace(0.0, 1.0, 7)


@pytest.mark.parametrize("phi", phis)
def test_rotation_heisenberg(phi):
    """ops: Tests the Heisenberg representation of the Rotation gate."""
    matrix = cv.Rotation._heisenberg_rep([phi])
    true_matrix = np.array(
        [[1, 0, 0], [0, np.cos(phi), -np.sin(phi)], [0, np.sin(phi), np.cos(phi)]]
    )
    assert np.allclose(matrix, true_matrix)


@pytest.mark.parametrize("phi", phis)
@pytest.mark.parametrize("mag", mags)
def test_squeezing_heisenberg(phi, mag):
    """ops: Tests the Heisenberg representation of the Squeezing gate."""
    r = mag
    matrix = cv.Squeezing._heisenberg_rep([r, phi])
    true_matrix = np.array(
        [
            [1, 0, 0],
            [0, np.cosh(r) - np.cos(phi) * np.sinh(r), -np.sin(phi) * np.sinh(r)],
            [0, -np.sin(phi) * np.sinh(r), np.cosh(r) + np.cos(phi) * np.sinh(r)],
        ]
    )
    assert np.allclose(matrix, true_matrix)


@pytest.mark.parametrize("phi", phis)
@pytest.mark.parametrize("mag", mags)
def test_displacement_heisenberg(phi, mag):
    """ops: Tests the Heisenberg representation of the Displacement gate."""
    r = mag
    hbar = 2
    matrix = cv.Displacement._heisenberg_rep([r, phi])
    true_matrix = np.array(
        [
            [1, 0, 0],
            [np.sqrt(2 * hbar) * r * np.cos(phi), 1, 0],
            [np.sqrt(2 * hbar) * r * np.sin(phi), 0, 1],
        ]
    )
    assert np.allclose(matrix, true_matrix)


@pytest.mark.parametrize("phi", phis)
@pytest.mark.parametrize("theta", phis)
def test_beamsplitter_heisenberg(phi, theta):
    """ops: Tests the Heisenberg representation of the Beamsplitter gate."""
    matrix = cv.Beamsplitter._heisenberg_rep([theta, phi])
    true_matrix = np.array(
        [
            [1, 0, 0, 0, 0],
            [
                0,
                np.cos(theta),
                0,
                -np.cos(phi) * np.sin(theta),
                -np.sin(phi) * np.sin(theta),
            ],
            [
                0,
                0,
                np.cos(theta),
                np.sin(phi) * np.sin(theta),
                -np.cos(phi) * np.sin(theta),
            ],
            [
                0,
                np.cos(phi) * np.sin(theta),
                -np.sin(phi) * np.sin(theta),
                np.cos(theta),
                0,
            ],
            [
                0,
                np.sin(phi) * np.sin(theta),
                np.cos(phi) * np.sin(theta),
                0,
                np.cos(theta),
            ],
        ]
    )
    assert np.allclose(matrix, true_matrix)


@pytest.mark.parametrize("phi", phis)
@pytest.mark.parametrize("mag", mags)
def test_two_mode_squeezing_heisenberg(phi, mag):
    """ops: Tests the Heisenberg representation of the Beamsplitter gate."""
    r = mag
    matrix = cv.TwoModeSqueezing._heisenberg_rep([r, phi])
    true_matrix = np.array(
        [
            [1, 0, 0, 0, 0],
            [0, np.cosh(r), 0, np.cos(phi) * np.sinh(r), np.sin(phi) * np.sinh(r)],
            [0, 0, np.cosh(r), np.sin(phi) * np.sinh(r), -np.cos(phi) * np.sinh(r)],
            [0, np.cos(phi) * np.sinh(r), np.sin(phi) * np.sinh(r), np.cosh(r), 0],
            [0, np.sin(phi) * np.sinh(r), -np.cos(phi) * np.sinh(r), 0, np.cosh(r)],
        ]
    )
    assert np.allclose(matrix, true_matrix)


@pytest.mark.parametrize("s", s_vals)
def test_quadratic_phase_heisenberg(s):
    """ops: Tests the Heisenberg representation of the QuadraticPhase gate."""
    matrix = cv.QuadraticPhase._heisenberg_rep([s])
    true_matrix = np.array([[1, 0, 0], [0, 1, 0], [0, s, 1]])
    assert np.allclose(matrix, true_matrix)


@pytest.mark.parametrize("s", s_vals)
def test_controlled_addition_heisenberg(s):
    """ops: Tests the Heisenberg representation of ControlledAddition gate.
    """
    matrix = cv.ControlledAddition._heisenberg_rep([s])
    true_matrix = np.array(
        [
            [1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0],
            [0, 0, 1, 0, -s],
            [0, s, 0, 1, 0],
            [0, 0, 0, 0, 1],
        ]
    )
    assert np.allclose(matrix, true_matrix)


@pytest.mark.parametrize("s", s_vals)
def test_controlled_phase_heisenberg(s):
    """Tests the Heisenberg representation of the ControlledPhase gate."""
    matrix = cv.ControlledPhase._heisenberg_rep([s])
    true_matrix = np.array(
        [
            [1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0],
            [0, 0, 1, s, 0],
            [0, 0, 0, 1, 0],
            [0, s, 0, 0, 1],
        ]
    )
    assert np.allclose(matrix, true_matrix)


class TestNonGaussian:
    """Tests that non-Gaussian gates are properly handled."""

    @pytest.mark.parametrize("gate", [cv.Kerr, cv.CrossKerr, cv.CubicPhase])
    def test_heisenberg_rep_nonguassian(self, gate):
        """ops: Tests that the `_heisenberg_rep` for a non-Gaussian gates is
        None
        """
        assert gate._heisenberg_rep(*[0.1] * gate.num_params) is None

    def test_heisenberg_transformation_nongaussian(self):
        """ops: Tests that proper exceptions are raised if we try to call the
        Heisenberg transformation of non-Gaussian gates."""
        op = cv.Kerr
        with pytest.raises(RuntimeError, match=r"not a Gaussian operation"):
            op_ = op(*[0.1] * op.num_params, wires=[0] * op.num_wires)
            op_.heisenberg_tr(op.num_wires)

        op = cv.CrossKerr
        with pytest.raises(ValueError):
            op_ = op(*[0.1] * op.num_params, wires=[0] * op.num_wires)
            op_.heisenberg_tr(op.num_wires)

        op = cv.CubicPhase
        with pytest.raises(RuntimeError):
            op_ = op(*[0.1] * op.num_params, wires=[0] * op.num_wires)
            op_.heisenberg_tr(op.num_wires)



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane.interface.torch` QNode interface.
"""

import pytest

import numpy as np

try:
    import torch
    from torch.autograd import Variable
except ImportError as e:
    pass

import pennylane as qml

from pennylane.qnode import _flatten, unflatten, QNode, QuantumFunctionError
from pennylane.plugins.default_qubit import CNOT, Rotx, Roty, Rotz, I, Y, Z
from pennylane._device import DeviceError


def expZ(state):
    return np.abs(state[0]) ** 2 - np.abs(state[1]) ** 2

@pytest.fixture()
def skip_if_no_torch_support(torch_support):
    if not torch_support:
        pytest.skip("Skipped, no torch support")   


@pytest.mark.usefixtures("skip_if_no_torch_support")
class TestTorchQNodeExceptions():
    """TorchQNode basic tests."""

    def test_qnode_fails_on_wrong_return_type(self, qubit_device_2_wires):
        """The qfunc must return only Expectations"""
        @qml.qnode(qubit_device_2_wires, interface='torch')
        def qf(x):
            qml.RX(x, wires=[0])
            return qml.expval(qml.PauliZ(0)), 0.3

        with pytest.raises(QuantumFunctionError, match='must return either'):
            qf(torch.tensor(0.5))

    def test_qnode_fails_on_expval_not_returned(self, qubit_device_2_wires):
        """All expectation values in the qfunc must be returned"""

        @qml.qnode(qubit_device_2_wires, interface='torch')
        def qf(x):
            qml.RX(x, wires=[0])
            ex = qml.expval(qml.PauliZ(1))
            return qml.expval(qml.PauliZ(0))

        with pytest.raises(QuantumFunctionError, match='All measured observables'):
            qf(torch.tensor(0.5))

    def test_qnode_fails_on_wrong_expval_order(self, qubit_device_2_wires):
        """Expvals must be returned in the order they were created in"""

        @qml.qnode(qubit_device_2_wires, interface='torch')
        def qf(x):
            qml.RX(x, wires=[0])
            ex = qml.expval(qml.PauliZ(1))
            return qml.expval(qml.PauliZ(0)), ex

        with pytest.raises(QuantumFunctionError, match='All measured observables'):
            qf(torch.tensor(0.5))

    def test_qnode_fails_on_gates_after_measurements(self, qubit_device_2_wires):
        """Gates have to precede measurements"""

        @qml.qnode(qubit_device_2_wires, interface='torch')
        def qf(x):
            qml.RX(x, wires=[0])
            ev = qml.expval(qml.PauliZ(1))
            qml.RY(0.5, wires=[0])
            return ev

        with pytest.raises(QuantumFunctionError, match='gates must precede'):
            qf(torch.tensor(0.5))

    def test_qnode_fails_on_multiple_measurements_of_same_wire(self, qubit_device_2_wires):
        """A wire can only be measured once"""
        
        @qml.qnode(qubit_device_2_wires, interface='torch')
        def qf(x):
            qml.RX(x, wires=[0])
            qml.CNOT(wires=[0, 1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1)), qml.expval(qml.PauliX(0))

        with pytest.raises(QuantumFunctionError, match='can only be measured once'):
            qf(torch.tensor(0.5))

    def test_qnode_fails_on_qfunc_with_too_many_wires(self, qubit_device_2_wires):
        """The device must have sufficient wires for the qfunc"""

        @qml.qnode(qubit_device_2_wires, interface='torch')
        def qf(x):
            qml.RX(x, wires=[0])
            qml.CNOT(wires=[0, 2])
            return qml.expval(qml.PauliZ(0))

        with pytest.raises(QuantumFunctionError, match='applied to invalid wire'):
            qf(torch.tensor(0.5))

    def test_qnode_fails_on_combination_of_cv_and_qbit_ops(self, qubit_device_1_wire):
        """CV and discrete operations must not be mixed"""
        
        @qml.qnode(qubit_device_1_wire, interface='torch')
        def qf(x):
            qml.RX(x, wires=[0])
            qml.Displacement(0.5, 0, wires=[0])
            return qml.expval(qml.PauliZ(0))

        with pytest.raises(QuantumFunctionError, match='Continuous and discrete'):
            qf(torch.tensor(0.5))

    def test_qnode_fails_for_cv_ops_on_qubit_device(self, qubit_device_1_wire):
        """A qubit device cannot execute CV operations"""

        @qml.qnode(qubit_device_1_wire, interface='torch')
        def qf(x):
            qml.Displacement(0.5, 0, wires=[0])
            return qml.expval(qml.X(0))

        with pytest.raises(DeviceError, match='Gate [a-zA-Z]+ not supported on device'):
            qf(torch.tensor(0.5))

    def test_qnode_fails_for_cv_observables_on_qubit_device(self, qubit_device_1_wire):
        """A qubit device cannot measure CV observables"""

        @qml.qnode(qubit_device_1_wire, interface='torch')
        def qf(x):
            return qml.expval(qml.X(0))

        with pytest.raises(DeviceError, match='Observable [a-zA-Z]+ not supported on device'):
            qf(torch.tensor(0.5))


@pytest.mark.usefixtures("skip_if_no_torch_support")
class TestTorchQNodeParameterHandling:
    """Test that the TorchQNode properly handles the parameters of qfuncs"""

    def test_qnode_fanout(self, qubit_device_1_wire, tol):
        """Tests that qnodes can compute the correct function when the same parameter is used in multiple gates."""

        @qml.qnode(qubit_device_1_wire, interface='torch')
        def circuit(reused_param, other_param):
            qml.RX(reused_param, wires=[0])
            qml.RZ(other_param, wires=[0])
            qml.RX(reused_param, wires=[0])
            return qml.expval(qml.PauliZ(0))

        thetas = torch.linspace(-2*np.pi, 2*np.pi, 7)

        for reused_param in thetas:
            for theta in thetas:
                other_param = theta ** 2 / 11
                y_eval = circuit(reused_param, other_param)
                Rx = Rotx(reused_param.numpy())
                Rz = Rotz(other_param.numpy())
                zero_state = np.array([1.,0.])
                final_state = (Rx @ Rz @ Rx @ zero_state)
                y_true = expZ(final_state)

                assert np.allclose(y_eval, y_true, atol=tol, rtol=0)

    def test_qnode_array_parameters_scalar_return(self, qubit_device_1_wire, tol):
        """Test that QNode can take arrays as input arguments, and that they interact properly with PyTorch.
           Test case for a circuit that returns a scalar."""

        # The objective of this test is not to check if the results are correctly calculated, 
        # but to check that the interoperability of the different return types works.
        @qml.qnode(qubit_device_1_wire, interface='torch')
        def circuit(dummy1, array, dummy2):
            qml.RY(0.5 * array[0,1], wires=0)
            qml.RY(-0.5 * array[1,1], wires=0)
            return qml.expval(qml.PauliX(0))  # returns a scalar

        grad_target = (np.array(1.), np.array([[0.5,  0.43879, 0], [0, -0.43879, 0]]), np.array(-0.4))
        cost_target = 1.03257

        args = (torch.tensor(0.46), torch.tensor([[2., 3., 0.3], [7., 4., 2.1]]), torch.tensor(-0.13))
        for i in args:
            i.requires_grad = True

        def cost(x, array, y):
            c = torch.as_tensor(circuit(torch.tensor(0.111), array, torch.tensor(4.5)), dtype=torch.float32)
            return c +0.5*array[0,0] +x -0.4*y

        cost_res = cost(*args)
        cost_res.backward()

        assert np.allclose(cost_res.detach().numpy(), cost_target, atol=tol, rtol=0)

        for i in range(3):
            assert np.allclose(args[i].grad.detach().numpy(), grad_target[i], atol=tol, rtol=0)

    def test_qnode_array_parameters_1_vector_return(self, qubit_device_1_wire, tol):
        """Test that QNode can take arrays as input arguments, and that they interact properly with PyTorch.
           Test case for a circuit that returns a 1-vector."""

        # The objective of this test is not to check if the results are correctly calculated, 
        # but to check that the interoperability of the different return types works.
        @qml.qnode(qubit_device_1_wire, interface='torch')
        def circuit(dummy1, array, dummy2):
            qml.RY(0.5 * array[0,1], wires=0)
            qml.RY(-0.5 * array[1,1], wires=0)
            return qml.expval(qml.PauliX(0)),  # note the comma, returns a 1-vector

        grad_target = (np.array(1.), np.array([[0.5,  0.43879, 0], [0, -0.43879, 0]]), np.array(-0.4))
        cost_target = 1.03257

        args = (torch.tensor(0.46), torch.tensor([[2., 3., 0.3], [7., 4., 2.1]]), torch.tensor(-0.13))
        for i in args:
            i.requires_grad = True

        def cost(x, array, y):
            c = torch.as_tensor(circuit(torch.tensor(0.111), array, torch.tensor(4.5)), dtype=torch.float32)
            c = c[0]  # get a scalar
            return c +0.5*array[0,0] +x -0.4*y

        cost_res = cost(*args)
        cost_res.backward()

        assert np.allclose(cost_res.detach().numpy(), cost_target, atol=tol, rtol=0)

        for i in range(3):
            assert np.allclose(args[i].grad.detach().numpy(), grad_target[i], atol=tol, rtol=0)

    def test_qnode_array_parameters_2_vector_return(self, qubit_device_2_wires, tol):
        """Test that QNode can take arrays as input arguments, and that they interact properly with PyTorch.
           Test case for a circuit that returns a 2-vector."""

        # The objective of this test is not to check if the results are correctly calculated, 
        # but to check that the interoperability of the different return types works.
        @qml.qnode(qubit_device_2_wires, interface='torch')
        def circuit(dummy1, array, dummy2):
            qml.RY(0.5 * array[0,1], wires=0)
            qml.RY(-0.5 * array[1,1], wires=0)
            qml.RY(array[1,0], wires=1)
            return qml.expval(qml.PauliX(0)), qml.expval(qml.PauliX(1))  # returns a 2-vector

        grad_target = (np.array(1.), np.array([[0.5,  0.43879, 0], [0, -0.43879, 0]]), np.array(-0.4))
        cost_target = 1.03257

        args = (torch.tensor(0.46), torch.tensor([[2., 3., 0.3], [7., 4., 2.1]]), torch.tensor(-0.13))
        for i in args:
            i.requires_grad = True

        def cost(x, array, y):
            c = torch.as_tensor(circuit(torch.tensor(0.111), array, torch.tensor(4.5)), dtype=torch.float32)
            c = c[0]  # get a scalar
            return c +0.5*array[0,0] +x -0.4*y

        cost_res = cost(*args)
        cost_res.backward()

        assert np.allclose(cost_res.detach().numpy(), cost_target, atol=tol, rtol=0)

        for i in range(3):
            assert np.allclose(args[i].grad.detach().numpy(), grad_target[i], atol=tol, rtol=0)


    def test_array_parameters_evaluate(self, qubit_device_2_wires, tol):
        """Test that array parameters gives same result as positional arguments."""
        a, b, c = torch.tensor(0.5), torch.tensor(0.54), torch.tensor(0.3)

        def ansatz(x, y, z):
            qml.QubitStateVector(np.array([1, 0, 1, 1])/np.sqrt(3), wires=[0, 1])
            qml.Rot(x, y, z, wires=0)
            qml.CNOT(wires=[0, 1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliY(1))

        @qml.qnode(qubit_device_2_wires, interface='torch')
        def circuit1(x, y, z):
            return ansatz(x, y, z)

        @qml.qnode(qubit_device_2_wires, interface='torch')
        def circuit2(x, array):
            return ansatz(x, array[0], array[1])

        @qml.qnode(qubit_device_2_wires, interface='torch')
        def circuit3(array):
            return ansatz(*array)

        positional_res = circuit1(a, b, c)
        array_res1 = circuit2(a, torch.tensor([b, c]))
        array_res2 = circuit3(torch.tensor([a, b, c]))

        assert np.allclose(positional_res.numpy(), array_res1.numpy(), atol=tol, rtol=0)
        assert np.allclose(positional_res.numpy(), array_res2.numpy(), atol=tol, rtol=0)

    def test_multiple_expectation_different_wires(self, qubit_device_2_wires, tol):
        """Tests that qnodes return multiple expectation values."""
        a, b, c = torch.tensor(0.5), torch.tensor(0.54), torch.tensor(0.3)

        @qml.qnode(qubit_device_2_wires, interface='torch')
        def circuit(x, y, z):
            qml.RX(x, wires=[0])
            qml.RZ(y, wires=[0])
            qml.CNOT(wires=[0, 1])
            qml.RY(y, wires=[0])
            qml.RX(z, wires=[0])
            return qml.expval(qml.PauliY(0)), qml.expval(qml.PauliZ(1))

        res = circuit(a, b, c)

        out_state = np.kron(Rotx(c.numpy()), I) @ np.kron(Roty(b.numpy()), I) @ CNOT \
            @ np.kron(Rotz(b.numpy()), I) @ np.kron(Rotx(a.numpy()), I) @ np.array([1, 0, 0, 0])

        ex0 = np.vdot(out_state, np.kron(Y, I) @ out_state)
        ex1 = np.vdot(out_state, np.kron(I, Z) @ out_state)
        ex = np.array([ex0, ex1])

        assert np.allclose(ex, res.numpy(), atol=tol, rtol=0)

    def test_multiple_keywordargs_used(self, qubit_device_2_wires, tol):
        """Tests that qnodes use multiple keyword arguments."""

        @qml.qnode(qubit_device_2_wires, interface='torch')
        def circuit(w, x=None, y=None):
            qml.RX(x, wires=[0])
            qml.RX(y, wires=[1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1))

        c = circuit(torch.tensor(1.), x=np.pi, y=np.pi)

        assert np.allclose(c.numpy(), [-1., -1.], atol=tol, rtol=0)

    def test_multidimensional_keywordargs_used(self, qubit_device_2_wires, tol):
        """Tests that qnodes use multi-dimensional keyword arguments."""
        def circuit(w, x=None):
            qml.RX(x[0], wires=[0])
            qml.RX(x[1], wires=[1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1))

        circuit = qml.QNode(circuit, qubit_device_2_wires).to_torch()

        c = circuit(torch.tensor(1.), x=[np.pi, np.pi])
        assert np.allclose(c.numpy(), [-1., -1.], atol=tol, rtol=0)

    def test_keywordargs_for_wires(self, qubit_device_2_wires, tol):
        """Tests that wires can be passed as keyword arguments."""
        default_q = 0

        def circuit(x, q=default_q):
            qml.RY(x, wires=0)
            return qml.expval(qml.PauliZ(q))

        circuit = qml.QNode(circuit, qubit_device_2_wires).to_torch()

        c = circuit(torch.tensor(np.pi), q=1)
        assert np.allclose(c, 1., atol=tol, rtol=0)

        c = circuit(torch.tensor(np.pi))
        assert np.allclose(c.numpy(), -1., atol=tol, rtol=0)

    def test_keywordargs_used(self, qubit_device_1_wire, tol):
        """Tests that qnodes use keyword arguments."""

        def circuit(w, x=None):
            qml.RX(x, wires=[0])
            return qml.expval(qml.PauliZ(0))

        circuit = qml.QNode(circuit, qubit_device_1_wire).to_torch()

        c = circuit(torch.tensor(1.), x=np.pi)
        assert np.allclose(c.numpy(), -1., atol=tol, rtol=0)

    def test_mixture_numpy_tensors(self, qubit_device_2_wires, tol):
        """Tests that qnodes work with python types and tensors."""

        @qml.qnode(qubit_device_2_wires, interface='torch')
        def circuit(w, x, y):
            qml.RX(x, wires=[0])
            qml.RX(y, wires=[1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1))

        c = circuit(torch.tensor(1.), np.pi, np.pi).detach().numpy()
        assert np.allclose(c, [-1., -1.], atol=tol, rtol=0)

    def test_keywordarg_updated_in_multiple_calls(self, qubit_device_2_wires):
        """Tests that qnodes update keyword arguments in consecutive calls."""

        def circuit(w, x=None):
            qml.RX(w, wires=[0])
            qml.RX(x, wires=[1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1))

        circuit = qml.QNode(circuit, qubit_device_2_wires).to_torch()

        c1 = circuit(torch.tensor(0.1), x=torch.tensor(0.))
        c2 = circuit(torch.tensor(0.1), x=np.pi)
        assert c1[1] != c2[1]

    def test_keywordarg_passes_through_classicalnode(self, qubit_device_2_wires, tol):
        """Tests that qnodes' keyword arguments pass through classical nodes."""

        def circuit(w, x=None):
            qml.RX(w, wires=[0])
            qml.RX(x, wires=[1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1))

        circuit = qml.QNode(circuit, qubit_device_2_wires).to_torch()

        def classnode(w, x=None):
            return circuit(w, x=x)

        c = classnode(torch.tensor(0.), x=np.pi)
        assert np.allclose(c.numpy(), [1., -1.], atol=tol, rtol=0)

    def test_keywordarg_gradient(self, qubit_device_2_wires, tol):
        """Tests that qnodes' keyword arguments work with gradients"""

        def circuit(x, y, input_state=np.array([0, 0])):
            qml.BasisState(input_state, wires=[0, 1])
            qml.RX(x, wires=[0])
            qml.RY(y, wires=[0])
            return qml.expval(qml.PauliZ(0))

        circuit = qml.QNode(circuit, qubit_device_2_wires).to_torch()

        x = 0.543
        y = 0.45632

        x_t = torch.autograd.Variable(torch.tensor(x), requires_grad=True)
        y_t = torch.autograd.Variable(torch.tensor(y), requires_grad=True)
        c = circuit(x_t, y_t, input_state=np.array([0, 0]))
        c.backward()
        assert np.allclose(x_t.grad.numpy(), [-np.sin(x)*np.cos(y)], atol=tol, rtol=0)
        assert np.allclose(y_t.grad.numpy(), [-np.sin(y)*np.cos(x)], atol=tol, rtol=0)

        x_t = torch.autograd.Variable(torch.tensor(x), requires_grad=True)
        y_t = torch.autograd.Variable(torch.tensor(y), requires_grad=True)
        c = circuit(x_t, y_t, input_state=np.array([1, 0]))
        c.backward()
        assert np.allclose(x_t.grad.numpy(), [np.sin(x)*np.cos(y)], atol=tol, rtol=0)
        assert np.allclose(y_t.grad.numpy(), [np.sin(y)*np.cos(x)], atol=tol, rtol=0)

        x_t = torch.autograd.Variable(torch.tensor(x), requires_grad=True)
        y_t = torch.autograd.Variable(torch.tensor(y), requires_grad=True)
        c = circuit(x_t, y_t)
        c.backward()
        assert np.allclose(x_t.grad.numpy(), [-np.sin(x)*np.cos(y)], atol=tol, rtol=0)
        assert np.allclose(y_t.grad.numpy(), [-np.sin(y)*np.cos(x)], atol=tol, rtol=0)


@pytest.mark.usefixtures("skip_if_no_torch_support")
class TestIntegration():
    """Integration tests to ensure the Torch QNode agrees with the NumPy QNode"""

    def test_qnode_evaluation_agrees(self, qubit_device_2_wires, tol):
        """Tests that simple example is consistent."""

        @qml.qnode(qubit_device_2_wires, interface='autograd')
        def circuit(phi, theta):
            qml.RX(phi[0], wires=0)
            qml.RY(phi[1], wires=1)
            qml.CNOT(wires=[0, 1])
            qml.PhaseShift(theta[0], wires=0)
            return qml.expval(qml.PauliZ(0))

        @qml.qnode(qubit_device_2_wires, interface='torch')
        def circuit_torch(phi, theta):
            qml.RX(phi[0], wires=0)
            qml.RY(phi[1], wires=1)
            qml.CNOT(wires=[0, 1])
            qml.PhaseShift(theta[0], wires=0)
            return qml.expval(qml.PauliZ(0))

        phi = [0.5, 0.1]
        theta = [0.2]

        phi_t = torch.tensor(phi)
        theta_t = torch.tensor(theta)

        autograd_eval = circuit(phi, theta)
        torch_eval = circuit_torch(phi_t, theta_t)
        assert np.allclose(autograd_eval, torch_eval.detach().numpy(), atol=tol, rtol=0)

    def test_qnode_gradient_agrees(self, qubit_device_2_wires, tol):
        """Tests that simple gradient example is consistent."""

        @qml.qnode(qubit_device_2_wires, interface='autograd')
        def circuit(phi, theta):
            qml.RX(phi[0], wires=0)
            qml.RY(phi[1], wires=1)
            qml.CNOT(wires=[0, 1])
            qml.PhaseShift(theta[0], wires=0)
            return qml.expval(qml.PauliZ(0))

        @qml.qnode(qubit_device_2_wires, interface='torch')
        def circuit_torch(phi, theta):
            qml.RX(phi[0], wires=0)
            qml.RY(phi[1], wires=1)
            qml.CNOT(wires=[0, 1])
            qml.PhaseShift(theta[0], wires=0)
            return qml.expval(qml.PauliZ(0))

        phi = [0.5, 0.1]
        theta = [0.2]

        phi_t = torch.autograd.Variable(torch.tensor(phi), requires_grad=True)
        theta_t = torch.autograd.Variable(torch.tensor(theta), requires_grad=True)

        dcircuit = qml.grad(circuit, [0, 1])
        autograd_grad = dcircuit(phi, theta)

        torch_eval = circuit_torch(phi_t, theta_t)
        torch_eval.backward()

        assert np.allclose(autograd_grad[0], phi_t.grad.detach().numpy(), atol=tol, rtol=0)
        assert np.allclose(autograd_grad[1], theta_t.grad.detach().numpy(), atol=tol, rtol=0)


gradient_test_data = [
    (0.5, -0.1),
    (0.0, np.pi),
    (-3.6, -3.6),
    (1.0, 2.5),
]


@pytest.mark.usefixtures("skip_if_no_torch_support")
class TestTorchGradients:
    """Integration tests involving gradients of QNodes and hybrid computations using the torch interface"""

    @pytest.fixture
    def qnodes(self):
        """Two QNodes to be used for the gradient tests"""
        dev = qml.device("default.qubit", wires=2)

        @qml.qnode(dev, interface="torch")
        def f(x):
            qml.RX(x, wires=0)
            return qml.expval(qml.PauliZ(0))

        @qml.qnode(dev, interface="torch")
        def g(y):
            qml.RY(y, wires=0)
            return qml.expval(qml.PauliX(0))

        return f, g

    @pytest.mark.parametrize("x, y", gradient_test_data)
    def test_addition_qnodes_gradient(self, qnodes, x, y):
        """Test the gradient of addition of two QNode circuits"""
        f, g = qnodes

        def add(a, b):
            return a + b

        xt = torch.autograd.Variable(torch.tensor(x), requires_grad=True)
        yt = torch.autograd.Variable(torch.tensor(y), requires_grad=True)

        # addition
        a = f(xt)
        b = g(yt)
        a.retain_grad()
        b.retain_grad()

        add(a, b).backward()
        assert a.grad == 1.0
        assert b.grad == 1.0

        # same tensor added to itself

        a = f(xt)
        a.retain_grad()

        add(a, a).backward()
        assert a.grad == 2.0

    @pytest.mark.parametrize("x, y", gradient_test_data)
    def test_subtraction_qnodes_gradient(self, qnodes, x, y):
        """Test the gradient of subtraction of two QNode circuits"""
        f, g = qnodes

        def subtract(a, b):
            return a - b

        xt = torch.autograd.Variable(torch.tensor(x), requires_grad=True)
        yt = torch.autograd.Variable(torch.tensor(y), requires_grad=True)

        # subtraction
        a = f(xt)
        b = g(yt)
        a.retain_grad()
        b.retain_grad()

        subtract(a, b).backward()
        assert a.grad == 1.0
        assert b.grad == -1.0

    @pytest.mark.parametrize("x, y", gradient_test_data)
    def test_multiplication_qnodes_gradient(self, qnodes, x, y):
        """Test the gradient of multiplication of two QNode circuits"""
        f, g = qnodes

        def mult(a, b):
            return a * b

        xt = torch.autograd.Variable(torch.tensor(x), requires_grad=True)
        yt = torch.autograd.Variable(torch.tensor(y), requires_grad=True)

        # multiplication
        a = f(xt)
        b = g(yt)
        a.retain_grad()
        b.retain_grad()

        mult(a, b).backward()
        assert a.grad == b
        assert b.grad == a

        a = f(xt)
        b = g(yt)
        a.retain_grad()
        b.retain_grad()

    @pytest.mark.parametrize("x, y", gradient_test_data)
    def test_division_qnodes_gradient(self, qnodes, x, y):
        """Test the gradient of division of two QNode circuits"""
        f, g = qnodes

        def div(a, b):
            return a / b

        xt = torch.autograd.Variable(torch.tensor(x), requires_grad=True)
        yt = torch.autograd.Variable(torch.tensor(y), requires_grad=True)

        # division
        # multiplication
        a = f(xt)
        b = g(yt)
        a.retain_grad()
        b.retain_grad()

        div(a, b).backward()
        assert a.grad == 1 / b
        assert b.grad == -a / b ** 2

    @pytest.mark.parametrize("x, y", gradient_test_data)
    def test_composition_qnodes_gradient(self, qnodes, x, y):
        """Test the gradient of composition of two QNode circuits"""
        f, g = qnodes

        def compose(f, x):
            return f(x)

        xt = torch.autograd.Variable(torch.tensor(x), requires_grad=True)
        yt = torch.autograd.Variable(torch.tensor(y), requires_grad=True)

        # compose function with xt as input
        compose(f, xt).backward()
        grad1 = xt.grad.detach().numpy()

        f(xt).backward()
        grad2 = xt.grad.detach().numpy()
        assert grad1 == grad2

        # compose function with a as input
        a = f(xt)
        a.retain_grad()

        compose(f, a).backward()
        grad1 = a.grad.detach().numpy()

        a = f(xt)
        a.retain_grad()

        f(a).backward()
        grad2 = a.grad.detach().numpy()
        assert grad1 == grad2

        # compose function with b as input
        b = g(yt)
        b.retain_grad()

        compose(f, b).backward()
        grad1 = b.grad.detach().numpy()

        b = g(yt)
        b.retain_grad()

        f(b).backward()
        grad2 = b.grad.detach().numpy()
        assert grad1 == grad2



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# PennyLane documentation build configuration file, created by
# sphinx-quickstart on Tue Apr 17 11:43:51 2018.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys, os, re

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('..'))
sys.path.insert(0, os.path.abspath('_ext'))
sys.path.insert(0, os.path.join(os.path.dirname(os.path.abspath('.')), 'doc'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '1.6'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.viewcode',
    'sphinxcontrib.bibtex',
    'edit_on_github',
    # 'sphinx_gallery.gen_gallery',
    "sphinx.ext.intersphinx",
    "sphinx_automodapi.automodapi",
]

autosummary_generate = True
autosummary_imported_members = True
automodapi_toctreedirnm = "code/api"


intersphinx_mapping = {"https://pennylane.ai/qml/": None}
mathjax_path = "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.5/MathJax.js?config=TeX-MML-AM_CHTML"

from glob import glob
import shutil
import os
import warnings

# sphinx_gallery_conf = {
#     # path to your example scripts
#     'examples_dirs': '../examples',
#     # path where to save gallery generated examples
#     'gallery_dirs': 'tutorials',
#     # build files that start 'pennylane_run'
#     'filename_pattern': r'pennylane_run',
#     # first notebook cell in generated Jupyter notebooks
#     'first_notebook_cell': "%matplotlib inline",
#     # thumbnail size
#     'thumbnail_size': (400, 400),
# }

# Remove warnings that occur when generating the the tutorials
warnings.filterwarnings("ignore", category=UserWarning, message=r"Matplotlib is currently using agg")
warnings.filterwarnings("ignore", category=FutureWarning, message=r"Passing \(type, 1\) or '1type' as a synonym of type is deprecated.+")
warnings.filterwarnings("ignore", category=UserWarning, message=r".+?Compilation using quilc will not be available\.")

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates', 'xanadu_theme']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'PennyLane'
copyright = """
    Ville Bergholm, Josh Izaac, Maria Schuld, Christian Gogolin, Carsten Blank, Keri McKiernan, and Nathan Killoran. <br>
PennyLane: Automatic differentiation of hybrid quantum-classical computations. arXiv:1811.04968, 2018.<br>
&copy; Copyright 2018-2019, Xanadu Quantum Technologies Inc."""
author = 'Xanadu Inc.'

add_module_names = False

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

import pennylane
# The full version, including alpha/beta/rc tags.
release = pennylane.__version__

# The short X.Y version.
version = re.match(r'^(\d+\.\d+)', release).expand(r'\1')

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# today_fmt is used as the format for a strftime call.
today_fmt = '%Y-%m-%d'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------


# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# html_theme = 'nature'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = '_static/favicon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
#html_sidebars = {
#    '**': [
#        'about.html',
#        'navigation.html',
#        'relations.html',  # needs 'show_related': True theme option to display
#        'searchbox.html',
#        'donate.html',
#    ]
#}
html_sidebars = {
    '**' : [
        'logo-text.html',
        'searchbox.html',
        'globaltoc.html',
        # 'sourcelink.html'
    ]
}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g., ".xhtml").
#html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'h', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'r', 'sv', 'tr'
#html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# Now only 'ja' uses this config value
#html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
#html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = 'PennyLanedoc'

# # -- Xanadu theme ---------------------------------------------------------
html_theme = 'xanadu_theme'
html_theme_path = ['.']

# Register the theme as an extension to generate a sitemap.xml
# extensions.append("guzzle_sphinx_theme")

# xanadu theme options (see theme.conf for more information)
html_theme_options = {

    # Set the path to a special layout to include for the homepage
    # "index_template": "special_index.html",

    # Set the name of the project to appear in the left sidebar.
    "project_nav_name": "PennyLane",

    # Set your Disqus short name to enable comments
    # "disqus_comments_shortname": "pennylane-1",

    # Set you GA account ID to enable tracking
    "google_analytics_account": "UA-130507810-2",

    # Path to a touch icon
    "touch_icon": "logo_new.png",

    # Specify a base_url used to generate sitemap.xml links. If not
    # specified, then no sitemap will be built.
    # "base_url": ""

    # Allow a separate homepage from the master_doc
    # "homepage": "index",

    # Allow the project link to be overriden to a custom URL.
    # "projectlink": "http://myproject.url",

    "large_toc": True,
    # colors
    "navigation_button": "#19b37b",
    "navigation_button_hover": "#0e714d",
    "toc_caption": "#19b37b",
    "toc_hover": "#19b37b",
    "table_header_bg": "#edf7f4",
    "table_header_border": "#19b37b",
    "download_button": "#19b37b",
    # gallery options
    # "github_repo": "XanaduAI/PennyLane",
    # "gallery_dirs": "tutorials",
}

edit_on_github_project = 'XanaduAI/pennylane'
edit_on_github_branch = 'master/doc'

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'PennyLanedoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

latex_additional_files = ['macros.tex']

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'PennyLane.tex', 'PennyLane Documentation',
     'Xanadu Inc.', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'pennylane', 'PennyLane Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'PennyLane', 'PennyLane Documentation',
     author, 'PennyLane', 'Xanadu quantum machine learning library.',
     'Miscellaneous'),
]


#============================================================

# the order in which autodoc lists the documented members
autodoc_member_order = 'bysource'

# inheritance_diagram graphviz attributes
inheritance_node_attrs = dict(color='lightskyblue1', style='filled')



r"""
.. _plugins_hybrid:

Plugins and Hybrid computation
==============================

This tutorial introduces the notion of hybrid computation by combining several PennyLane
plugins. We first introduce PennyLane's `Strawberry Fields plugin <https://pennylane-sf.readthedocs.io>`_
and use it to explore a non-Gaussian photonic circuit. We then combine this photonic circuit with a
qubit circuit — along with some classical processing — to create and optimize a fully hybrid computation.
Be sure to read through the introductory :ref:`qubit rotation <qubit_rotation>` and
:ref:`Gaussian transformation <gaussian_transformation>` tutorials before attempting this tutorial.

.. note::

    To follow along with this tutorial on your own computer, you will require the
    `PennyLane-SF plugin <https://pennylane-sf.readthedocs.io>`_, in order to access the
    `Strawberry Fields <https://strawberryfields.readthedocs.io>`_ Fock backend using
    PennyLane. It can be installed via pip:

    .. code-block:: bash

        pip install pennylane-sf

.. _photon_redirection:


A non-Gaussian circuit
----------------------

We first consider a photonic circuit which is similar in spirit to the
:ref:`qubit rotation <qubit_rotation>` circuit:

.. figure:: ../../examples/figures/photon_redirection.png
    :align: center
    :width: 30%
    :target: javascript:void(0);

Breaking this down, step-by-step:

1. **We start the computation with two qumode subsystems**. In PennyLane, we use the
   shorthand 'wires' to refer to quantum subsystems, whether they are qumodes, qubits, or
   any other kind of quantum register.

2. **Prepare the state** :math:`\left|1,0\right\rangle`. That is, the first wire (wire 0) is prepared
   in a single-photon state, while the second
   wire (wire 1) is prepared in the vacuum state. The former state is non-Gaussian,
   necessitating the use of the ``'strawberryfields.fock'`` backend device.

3. **Both wires are then incident on a beamsplitter**, with free parameters :math:`\theta`and :math:`\phi`.
   Here, we have the convention that the beamsplitter transmission amplitude is :math:`t=\cos\theta`,
   and the reflection amplitude is
   :math:`r=e^{i\phi}\sin\theta`. See :ref:`operations <intro_ref_ops_cv>` for a full list of operation conventions.

4. **Finally, we measure the mean photon number** :math:`\left\langle \hat{n}\right\rangle` of the second wire, where

   .. math:: \hat{n} = \ad\a

   is the number operator, acting on the Fock basis number states, such that :math:`\hat{n}\left|n\right\rangle = n\left|n\right\rangle`.

The aim of this tutorial is to optimize the beamsplitter parameters :math:`(\theta, \phi)` such
that the expected photon number of the second wire is **maximized**. Since the beamsplitter
is a passive optical element that preserves the total photon number, this to the output
state :math:`\left|0,1\right\rangle` — i.e., when the incident photon from the first wire has been
'redirected' to the second wire.

.. _photon_redirection_calc:

Exact calculation
~~~~~~~~~~~~~~~~~

To compare with later numerical results, we can first consider what happens analytically.
The initial state of the circuit is :math:`\left|\psi_0\right\rangle=\left|1,0\right\rangle`, and the output state
of the system is of the form :math:`\left|\psi\right\rangle = a\left|1, 0\right\rangle + b\left|0,1\right\rangle`, where
:math:`|a|^2+|b|^2=1`. We may thus write the output state as a vector in this
computational basis, :math:`\left|\psi\right\rangle = \begin{bmatrix}a & b\end{bmatrix}^T`.

The beamsplitter acts on this two-dimensional subspace as follows:

.. math::
    \left|\psi\right\rangle = B(\theta, \phi)\left|1, 0\right\rangle = \begin{bmatrix}
        \cos\theta & -e^{-i\phi}\sin\theta\\
        e^{i\phi}\sin\theta & \cos\theta
    \end{bmatrix}\begin{bmatrix} 1\\ 0\end{bmatrix} = \begin{bmatrix}
        \cos\theta\\
        e^{i\phi} \sin\theta
    \end{bmatrix}

Furthermore, the mean photon number of the second wire is

.. math::

    \left\langle{\hat{n}_1}\right\rangle = \langle{\psi}\mid{\hat{n}_1}\mid{\psi}\rangle = |e^{i\phi} \sin\theta|^2
    \langle{0,1}\mid{\hat{n}_1}\mid{0,1}\rangle = \sin^2 \theta.

Therefore, we can see that:

1. :math:`0\leq \left\langle \hat{n}_1\right\rangle\leq 1`: the output of the quantum circuit is
   bound between 0 and 1;

2. :math:`\frac{\partial}{\partial \phi} \left\langle \hat{n}_1\right\rangle=0`: the output of the
   quantum circuit is independent of the beamsplitter phase :math:`\phi`;

3. The output of the quantum circuit above is maximised when :math:`\theta=(2m+1)\pi/2`
   for :math:`m\in\mathbb{Z}_0`.
"""

##############################################################################
# Loading the plugin device
# -------------------------
#
# While PennyLane provides a basic qubit simulator (``'default.qubit'``) and a basic CV
# Gaussian simulator (``'default.gaussian'``), the true power of PennyLane comes from its
# :ref:`plugin ecosystem <plugins>`, allowing quantum computations to be run on a variety
# of quantum simulator and hardware devices.
#
# For this circuit, we will be using the ``'strawberryfields.fock'`` device to construct
# a QNode. This allows the underlying quantum computation to be performed using the
# `Strawberry Fields <https://strawberryfields.readthedocs.io>`_ Fock backend.
#
# As usual, we begin by importing PennyLane and the wrapped version of NumPy provided by PennyLane:

import pennylane as qml
from pennylane import numpy as np

##############################################################################
# Next, we create a device to run the quantum node. This is easy in PennyLane; as soon as
# the PennyLane-SF plugin is installed, the ``'strawberryfields.fock'`` device can be loaded
# — no additional commands or library imports required.

dev_fock = qml.device("strawberryfields.fock", wires=2, cutoff_dim=2)

##############################################################################
# Compared to the default devices provided with PennyLane, the ``'strawberryfields.fock'``
# device requires the additional keyword argument:
#
# * ``cutoff_dim``: the Fock space truncation used to perform the quantum simulation
#
# .. note::
#
#     Devices provided by external plugins may require additional arguments and keyword arguments
#     — consult the plugin documentation for more details.


##############################################################################
# Constructing the QNode
# ----------------------
#
# Now that we have initialized the device, we can construct our quantum node. Like
# the other tutorials, we use the :mod:`qnode decorator <pennylane.decorator>`
# to convert our quantum function (encoded by the circuit above) into a quantum node
# running on Strawberry Fields.


@qml.qnode(dev_fock)
def photon_redirection(params):
    qml.FockState(1, wires=0)
    qml.Beamsplitter(params[0], params[1], wires=[0, 1])
    return qml.expval(qml.NumberOperator(1))


##############################################################################
# The ``'strawberryfields.fock'`` device supports all CV objects provided by PennyLane;
# see :ref:`CV operations <intro_vcircuits>`.


##############################################################################
# Optimization
# ------------
#
# Let's now use one of the built-in PennyLane optimizers in order to
# carry out photon redirection. Since we wish to maximize the mean photon number of
# the second wire, we can define our cost function to minimize the *negative* of the circuit output.


def cost(params):
    return -photon_redirection(params)


##############################################################################
# To begin our optimization, let's choose the following small initial values of
# :math:`\theta` and :math:`\phi`:

init_params = np.array([0.01, 0.01])
print(cost(init_params))

##############################################################################
# Here, we choose the values of :math:`\theta` and :math:`\phi` to be very close to zero;
# this results in :math:`B(\theta,\phi)\approx I`, and the output of the quantum
# circuit will be very close to :math:`\left|1, 0\right\rangle` — i.e., the circuit leaves the photon in the first mode.
#
# Why don't we choose :math:`\theta=0` and :math:`\phi=0`?
#
# At this point in the parameter space, :math:`\left\langle \hat{n}_1\right\rangle = 0`, and
# :math:`\frac{d}{d\theta}\left\langle{\hat{n}_1}\right\rangle|_{\theta=0}=2\sin\theta\cos\theta|_{\theta=0}=0`.
# Since the gradient is zero at those initial parameter values, the optimization
# algorithm would never descend from the maximum.
#
# This can also be verified directly using PennyLane:

dphoton_redirection = qml.grad(photon_redirection, argnum=0)
print(dphoton_redirection([0.0, 0.0]))

##############################################################################
# Now, let's use the :class:`~.GradientDescentOptimizer`, and update the circuit
# parameters over 100 optimization steps.

# initialise the optimizer
opt = qml.GradientDescentOptimizer(stepsize=0.4)

# set the number of steps
steps = 100
# set the initial parameter values
params = init_params

for i in range(steps):
    # update the circuit parameters
    params = opt.step(cost, params)

    if (i + 1) % 5 == 0:
        print("Cost after step {:5d}: {: .7f}".format(i + 1, cost(params)))

print("Optimized rotation angles: {}".format(params))


##############################################################################
# Comparing this to the :ref:`exact calculation <photon_redirection_calc>` above,
# this is close to the optimum value of :math:`\theta=\pi/2`, while the value of
# :math:`\phi` has not changed — consistent with the fact that :math:`\left\langle \hat{n}_1\right\rangle`
# is independent of :math:`\phi`.
#
# .. _hybrid_computation_example:
#
# Hybrid computation
# ------------------
#
# To really highlight the capabilities of PennyLane, let's now combine the qubit-rotation QNode
# from the :ref:`qubit rotation tutorial <qubit_rotation>` with the CV photon-redirection
# QNode from above, as well as some classical processing, to produce a truly hybrid
# computational model.
#
# First, we define a computation consisting of three steps: two quantum nodes (the qubit rotation
# and photon redirection circuits, running on the ``'default.qubit'`` and
# ``'strawberryfields.fock'`` devices, respectively), along with a classical function, that simply
# returns the squared difference of its two inputs using NumPy:

# create the devices
dev_qubit = qml.device("default.qubit", wires=1)
dev_fock = qml.device("strawberryfields.fock", wires=2, cutoff_dim=10)


@qml.qnode(dev_qubit)
def qubit_rotation(phi1, phi2):
    """Qubit rotation QNode"""
    qml.RX(phi1, wires=0)
    qml.RY(phi2, wires=0)
    return qml.expval(qml.PauliZ(0))


@qml.qnode(dev_fock)
def photon_redirection(params):
    """The photon redirection QNode"""
    qml.FockState(1, wires=0)
    qml.Beamsplitter(params[0], params[1], wires=[0, 1])
    return qml.expval(qml.NumberOperator(1))


def squared_difference(x, y):
    """Classical node to compute the squared
    difference between two inputs"""
    return np.abs(x - y) ** 2


##############################################################################
# Now, we can define an objective function associated with the optimization, linking together
# our three subcomponents. Here, we wish to
# perform the following hybrid quantum-classical optimization:
#
# .. figure:: ../../examples/figures/hybrid_graph.png
#     :align: center
#     :width: 70%
#     :target: javascript:void(0);
#
# 1. The qubit-rotation circuit will contain fixed rotation angles :math:`\phi_1` and :math:`\phi_2`.
#
# 2. The photon-redirection circuit will contain two free parameters, the beamsplitter angles
#    :math:`\theta` and :math:`\phi`, which are to be optimized.
#
# 3. The outputs of both QNodes will then be fed into the classical node, returning the
#    squared difference of the two quantum functions.
#
# 4. Finally, the optimizer will calculate the gradient of the entire computation with
#    respect to the free parameters :math:`\theta` and :math:`\phi`, and update their values.
#
# In essence, we are optimizing the photon-redirection circuit to return the **same expectation value**
# as the qubit-rotation circuit, even though they are two completely independent quantum systems.
#
# We can translate this computational graph to the following function, which combines the three
# nodes into a single hybrid computation. Below, we choose default values
# :math:`\phi_1=0.5`, :math:`\phi_2=0.1`:


def cost(params, phi1=0.5, phi2=0.1):
    """Returns the squared difference between
    the photon-redirection and qubit-rotation QNodes, for
    fixed values of the qubit rotation angles phi1 and phi2"""
    qubit_result = qubit_rotation(phi1, phi2)
    photon_result = photon_redirection(params)
    return squared_difference(qubit_result, photon_result)


##############################################################################
# Now, we use the built-in :class:`~.GradientDescentOptimizer` to perform the optimization
# for 100 steps. As before, we choose initial beamsplitter parameters of
# :math:`\theta=0.01`, :math:`\phi=0.01`.

# initialise the optimizer
opt = qml.GradientDescentOptimizer(stepsize=0.4)

# set the number of steps
steps = 100
# set the initial parameter values
params = np.array([0.01, 0.01])

for i in range(steps):
    # update the circuit parameters
    params = opt.step(cost, params)

    if (i + 1) % 5 == 0:
        print("Cost after step {:5d}: {: .7f}".format(i + 1, cost(params)))

print("Optimized rotation angles: {}".format(params))

##############################################################################
# Substituting this into the photon redirection QNode shows that it now produces
# the same output as the qubit rotation QNode:

result = [1.20671364, 0.01]
print(photon_redirection(result))
print(qubit_rotation(0.5, 0.1))

##############################################################################
# This is just a simple example of the kind of hybrid computation that can be carried
# out in PennyLane. Quantum nodes (bound to different devices) and classical
# functions can be combined in many different and interesting ways.
